""" Interactive routines for EELS analysis

this file provides aditional dialogs for EELS quantification

Author: Gerd Duscher
"""

import numpy as np

from PyQt5 import QtWidgets, QtCore, QtGui
import sidpy
import matplotlib.patches as patches
from matplotlib.widgets import RectangleSelector, SpanSelector

import h5py  # TODO: needs to go
import matplotlib.pyplot as plt

from IPython.display import display
import ipywidgets as widgets

from pyTEMlib import eels_tools as eels
from pyTEMlib import eels_dialog

from pyTEMlib import info_dialog
from pyTEMlib import peak_dialog

major_edges = ['K1', 'L3', 'M5', 'N5']
all_edges = ['K1', 'L1', 'L2', 'L3', 'M1', 'M2', 'M3', 'M4', 'M5', 'N1', 'N2', 'N3', 'N4', 'N5', 'N6', 'N7', 'O1', 'O2',
             'O3', 'O4', 'O5', 'O6', 'O7', 'P1', 'P2', 'P3']
first_close_edges = ['K1', 'L3', 'M5', 'M3', 'N5', 'N3']

CompositionDialog = eels_dialog.EELSDialog
CurveVisualizer = eels_dialog.CurveVisualizer
InfoDialog = info_dialog.InfoDialog
PeakFitDialog = peak_dialog.PeakFitDialog


class PeriodicTableDialog(QtWidgets.QDialog):
    """ Modal dialog to get a selection of elements.

    Elements that are not having a valid cross sections are disabled.

    Parameters
    ----------
    initial_elements: list of str
        the elements that are already selected
    energy_scale: list or numpy array
        energy-scale of spectrum/spectra to determine likely edges

    Returns
    -------
    list of strings: elements.

    Example
    -------
    >> PT_dialog =  periodic_table_dialog(None, ['Mn', 'O'])
    >> if PT_dialog.exec_() == periodic_table_dialog.Accepted:
    >>     selected_elements = PT_dialog.get_output()
    >> print(selected_elements)
    """

    signal_selected = QtCore.pyqtSignal(list)

    def __init__(self, initial_elements=None, energy_scale=None, parent=None):
        super(PeriodicTableDialog, self).__init__(None, QtCore.Qt.WindowStaysOnTopHint)

        if initial_elements is None:
            initial_elements = [' ']
        if energy_scale is None:
            energy_scale = [100., 150., 200.]
        self.parent = parent
        self._output = []
        self.elements_selected = initial_elements
        self.energy_scale = np.array(energy_scale)

        self.setWindowTitle("Periodic Table")
        likely_edges = get_likely_edges(self.energy_scale)
        self.likely_edges = likely_edges

        # GD:font = wx.Font(10, wx.MODERN, wx.NORMAL, wx.BOLD)
        self.buttons1 = []
        self.button = []
        self.pt_info = get_periodic_table_info()
        self.init_ui()

        for button in self.button:
            if button.text() in initial_elements:
                button.toggle()
            pass

    def on_close(self):
        self.get_output()
        self.signal_selected[list].emit(self._output)
        self.accept()

    def get_output(self):
        self._output = []
        for btn in self.button:
            if btn.isChecked():
                self._output.append(btn.text())

    def exec_(self):
        super(PeriodicTableDialog, self).exec_()
        return self._output

    def init_ui(self):

        v_sizer = QtWidgets.QVBoxLayout()
        g_sizer = QtWidgets.QGridLayout()

        main_group = QtWidgets.QWidget()

        color1 = "background-color: lightblue;\n"
        color1l = "background-color: dodgerblue;\n"
        color2 = "background-color: coral;\n"

        for symbol, parameter in self.pt_info.items():
            self.button.append(QtWidgets.QPushButton(symbol))
            if parameter['PT_row'] > 7:
                self.button[-1].setStyleSheet(color2)
            elif '*' in symbol:
                self.button[-1].setStyleSheet(color2)
            else:
                if symbol in self.likely_edges:
                    self.button[-1].setStyleSheet(color1l)
                else:
                    self.button[-1].setStyleSheet(color1)
            if parameter['Z'] == 0:
                self.button[-1].setEnabled(False)
            self.button[-1].setFixedWidth(50)
            self.button[-1].setCheckable(True)
            g_sizer.addWidget(self.button[-1], parameter['PT_row'], parameter['PT_col'])
        main_group.setLayout(g_sizer)

        v_sizer.addWidget(main_group)
        self.setLayout(v_sizer)

        ok_button = QtWidgets.QPushButton('OK')
        ok_button.clicked.connect(self.on_close)

        v_sizer.addWidget(ok_button)
        self.setLayout(v_sizer)


class EnergySelector(QtWidgets.QDialog):
    """Dialog and cursor to set energy scale"""

    signal_selected = QtCore.pyqtSignal(bool)

    def __init__(self, dset=None):
        super(EnergySelector, self).__init__(None, QtCore.Qt.WindowStaysOnTopHint)

        if not isinstance(dset, sidpy.Dataset):
            return
        if dset is None:
            return
        if dset.view is None:
            return
        self.dataset = dset

        if hasattr(dset.view, 'axis'):
            self.axis = dset.view.axis
            # self.setWindowTitle('p')
        elif hasattr(dset.view, 'axes'):
            self.axis = dset.view.axes[1]
        else:
            return

        self.spec_dim = -1
        for dim, axis in self.dataset._axes.items():
            if axis.dimension_type == sidpy.DimensionType.SPECTRAL:
                self.spec_dim = dim
        if self.spec_dim < 0:
            raise TypeError('We need at least one SPECTRAL dimension')

        self.energy_scale = self.dataset._axes[self.spec_dim].values
        self.dispersion = self.energy_scale[1] - self.energy_scale[0]
        self.offset = self.energy_scale[0]
        self.spectrum = np.zeros(2)

        self.change = 0

        self.x_min = self.energy_scale[int(len(self.energy_scale)/4)]
        self.x_max = self.energy_scale[int(len(self.energy_scale) / 4*3)]
        self.setWindowTitle("Select Energy")

        valid_float = QtGui.QDoubleValidator()

        layout = QtWidgets.QGridLayout()
        layout.setVerticalSpacing(2)
        self.label1 = QtWidgets.QLabel('Start:')
        self.edit1 = QtWidgets.QLineEdit('0')
        self.edit1.setValidator(valid_float)
        self.unit1 = QtWidgets.QLabel('eV')

        self.label2 = QtWidgets.QLabel('End:')
        self.edit2 = QtWidgets.QLineEdit('0')
        self.edit2.setValidator(valid_float)
        self.unit2 = QtWidgets.QLabel('eV')

        self.label3 = QtWidgets.QLabel('Dispersion:')
        self.edit3 = QtWidgets.QLineEdit('0')
        self.edit3.setValidator(valid_float)
        self.unit3 = QtWidgets.QLabel('eV')

        self.edit1.editingFinished.connect(self.on_enter)
        self.edit2.editingFinished.connect(self.on_enter)
        self.edit3.editingFinished.connect(self.on_enter)

        layout.addWidget(self.label1, 0, 0)
        layout.addWidget(self.edit1, 0, 1)
        layout.addWidget(self.unit1, 0, 2)

        layout.addWidget(self.label2, 1, 0)
        layout.addWidget(self.edit2, 1, 1)
        layout.addWidget(self.unit2, 1, 2)

        layout.addWidget(self.label3, 2, 0)
        layout.addWidget(self.edit3, 2, 1)
        layout.addWidget(self.unit3, 2, 2)

        self.ok_button = QtWidgets.QPushButton('OK')
        self.ok_button.clicked.connect(self.on_close)
        self.cancel_button = QtWidgets.QPushButton('Cancel')
        self.cancel_button.clicked.connect(self.on_close)

        layout.addWidget(self.ok_button, 3, 0)
        layout.addWidget(self.cancel_button, 3, 2)

        self.setLayout(layout)
        self.edit1.setFocus()
        self.plot()

        self.selector = RangeSelector(self.axis, self.line_select_callback)
        self.edit1.setText(f'{self.x_min:.3f}')
        self.edit2.setText(f'{self.x_max:.3f}')
        self.edit3.setText(f'{self.dispersion:.4f}')
        self.update()

    def line_select_callback(self, eclick, erelease):
        y_min, y_max = self.axis.get_ylim()
        self.x_min = self.selector.extents[0]
        self.x_max = self.selector.extents[1]
        self.selector.extents = (self.x_min, self.x_max, y_min, y_max)

        self.edit1.setText(f'{self.x_min:.3f}')
        self.edit2.setText(f'{self.x_max:.3f}')

    def on_enter(self):
        sender = self.sender()

        if sender == self.edit1:
            value = float(str(sender.displayText()).strip())
            if value == self.x_min:
                return
            self.change = value - self.x_min
            self.x_min += self.change
            self.x_max += self.change
            self.offset += self.change

            self.edit1.setText(f"{self.x_min:.2f}")
            self.edit2.setText(f"{self.x_max:.2f}")

            self.energy_scale = np.arange(len(self.energy_scale)) * self.dispersion + self.offset

            self.update()
            # self.axis.draw()
            # self.setWindowTitle(f'shift, {self.change}, {self.x_min}')

        elif sender == self.edit2:
            value = float(str(sender.displayText()).strip())
            if value == self.x_max:
                return
            start_channel = np.searchsorted(self.energy_scale, self.x_min)
            end_channel = np.searchsorted(self.energy_scale, self.x_max)

            self.x_max = value

            if end_channel - start_channel != 0:
                self.dispersion = (self.x_max - self.x_min) / (end_channel - start_channel)
            self.offset = self.x_min - start_channel * self.dispersion
            self.edit2.setText(f"{self.x_max:.3f}")
            self.edit3.setText(f"{self.dispersion:.4f}")
            self.energy_scale = np.arange(len(self.energy_scale)) * self.dispersion + self.offset

            self.update()
            # self.axis.draw()
            # self.setWindowTitle(f'range, {self.change}, {self.dispersion}')

        elif sender == self.edit3:
            value = float(str(sender.displayText()).strip())
            if self.dispersion == value:
                return

            start_channel = np.searchsorted(self.energy_scale, self.x_min)
            end_channel = np.searchsorted(self.energy_scale, self.x_max)
            self.dispersion = value
            self.energy_scale = np.arange(len(self.energy_scale)) * self.dispersion + self.offset
            self.x_min = self.energy_scale[start_channel]
            self.x_max = self.energy_scale[end_channel]
            self.update()
            # self.axis.draw()
            self.edit3.setText(f"{self.dispersion:.3f}")
            self.change = 0

    def on_close(self):
        sender = self.sender()
        if sender == self.ok_button:
            pass
            self.dataset.set_dimension(self.spec_dim, sidpy.Dimension(self.energy_scale, name='energy_scale',
                                                                      units='eV', quantity='energy loss',
                                                                      dimension_type='spectral'))
        else:
            pass
        self.selector.set_active(False)
        self.signal_selected[bool].emit(True)
        self.accept()

    def plot(self):
        if self.dataset.data_type == sidpy.DataType.SPECTRAL_IMAGE:
            self.spectrum = self.dataset.view.get_spectrum()
        else:
            self.spectrum = np.array(self.dataset)
        x_limit = self.axis.get_xlim()
        y_limit = self.axis.get_ylim()

        self.axis.clear()
        self.cplot = self.axis.plot(self.energy_scale, self.spectrum, label='spectrum')
        self.axis.set_xlim(x_limit)
        self.axis.set_ylim(y_limit)

        self.axis.figure.canvas.draw()

    def update(self):
        x_limit = self.axis.get_xlim()
        y_limit = self.axis.get_ylim()
        self.selector.extents = (self.x_min, self.x_max, y_limit[0], y_limit[1])

        x_limit = np.array(x_limit) + self.change

        self.cplot[0].set_data(self.energy_scale, self.spectrum)
        self.axis.set_xlim(x_limit)
        self.axis.set_ylim(y_limit)
        self.axis.figure.canvas.draw()


class RegionSelector(object):
    """Selects fitting region and the regions that are excluded for each edge.

        Select a region with a spanSelector and then type 'a' for all of the fitting region or a number for the edge
        you want to define the region excluded from the fit (solid state effects).

        see Chapter4 'CH4-Working_with_X-Sections,ipynb' notebook

    """

    def __init__(self, ax):
        self.ax = ax
        self.regions = {}
        self.rect = None
        self.xmin = 0
        self.width = 0

        self.span = SpanSelector(ax, self.on_select1, 'horizontal', useblit=True,
                                 rectprops=dict(alpha=0.5, facecolor='red'), span_stays=True)
        self.cid = ax.figure.canvas.mpl_connect('key_press_event', self.click)
        self.draw = ax.figure.canvas.mpl_connect('draw_event', self.onresize)

    def on_select1(self, xmin, xmax):
        self.xmin = xmin
        self.width = xmax - xmin

    def onresize(self, event):
        self.update()

    def delete_region(self, key):
        if key in self.regions:
            if 'Rect' in self.regions[key]:
                self.regions[key]['Rect'].remove()
                self.regions[key]['Text'].remove()
            del (self.regions[key])

    def update(self):

        y_min, y_max = self.ax.get_ylim()
        for key in self.regions:
            if 'Rect' in self.regions[key]:
                self.regions[key]['Rect'].remove()
                self.regions[key]['Text'].remove()

            xmin = self.regions[key]['xmin']
            width = self.regions[key]['width']
            height = y_max - y_min
            alpha = self.regions[key]['alpha']
            color = self.regions[key]['color']
            self.regions[key]['Rect'] = patches.Rectangle((xmin, y_min), width, height,
                                                          edgecolor=color, alpha=alpha, facecolor=color)
            self.ax.add_patch(self.regions[key]['Rect'])

            self.regions[key]['Text'] = self.ax.text(xmin, y_max, self.regions[key]['text'], verticalalignment='top')

    def click(self, event):
        if str(event.key) in ['1', '2', '3', '4', '5', '6']:
            key = str(event.key)
            text = 'exclude \nedge ' + key
            alpha = 0.5
            color = 'red'
        elif str(event.key) in ['a', 'A', 'B', 'b', 'f', 'F']:
            key = '0'
            color = 'blue'
            alpha = 0.2
            text = 'fit region'
        else:
            return

        if key not in self.regions:
            self.regions[key] = {}

        self.regions[key]['xmin'] = self.xmin
        self.regions[key]['width'] = self.width
        self.regions[key]['color'] = color
        self.regions[key]['alpha'] = alpha
        self.regions[key]['text'] = text

        self.update()

    def set_regions(self, region, start_x, width):
        key = ''
        if 'fit' in str(region):
            key = '0'
        if region in ['0', '1', '2', '3', '4', '5', '6']:
            key = region
        if region in [0, 1, 2, 3, 4, 5, 6]:
            key = str(region)

        if key not in self.regions:
            self.regions[key] = {}
            if key in ['1', '2', '3', '4', '5', '6']:
                self.regions[key]['text'] = 'exclude \nedge ' + key
                self.regions[key]['alpha'] = 0.5
                self.regions[key]['color'] = 'red'
            elif key == '0':
                self.regions[key]['text'] = 'fit region'
                self.regions[key]['alpha'] = 0.2
                self.regions[key]['color'] = 'blue'

        self.regions[key]['xmin'] = start_x
        self.regions[key]['width'] = width

        self.update()

    def get_regions(self):
        tags = {}
        for key in self.regions:
            if key == '0':
                area = 'fit_area'
            else:
                area = key
            tags[area] = {}
            tags[area]['start_x'] = self.regions[key]['xmin']
            tags[area]['width_x'] = self.regions[key]['width']

        return tags

    def disconnect(self):
        for key in self.regions:
            if 'Rect' in self.regions[key]:
                self.regions[key]['Rect'].remove()
                self.regions[key]['Text'].remove()
        del self.span
        self.ax.figure.canvas.mpl_disconnect(self.cid)
        # self.ax.figure.canvas.mpl_disconnect(self.draw)
        pass


class RangeSelector(RectangleSelector):
    """Select ranges of edge fitting interactively"""
    def __init__(self, ax, on_select):
        drawtype = 'box'
        spancoords = 'data'
        rectprops = dict(facecolor="blue", edgecolor="black", alpha=0.2, fill=True)

        super().__init__(ax, on_select, drawtype=drawtype,
                         minspanx=0, minspany=0, useblit=False,
                         lineprops=None, rectprops=rectprops, spancoords=spancoords,
                         button=None, maxdist=10, marker_props=None,
                         interactive=True, state_modifier_keys=None)

        self.artists = [self.to_draw, self._center_handle.artist,
                        self._edge_handles.artist]

    def draw_shape(self, extents):
        x0, x1, y0, y1 = extents
        xmin, xmax = sorted([x0, x1])
        # ymin, ymax = sorted([y0, y1])
        xlim = sorted(self.ax.get_xlim())
        ylim = sorted(self.ax.get_ylim())

        xmin = max(xlim[0], xmin)
        ymin = ylim[0]
        xmax = min(xmax, xlim[1])
        ymax = ylim[1]

        self.to_draw.set_x(xmin)
        self.to_draw.set_y(ymin)
        self.to_draw.set_width(xmax - xmin)
        self.to_draw.set_height(ymax - ymin)


def get_likely_edges(energy_scale):
    """get likely ionization edges within energy_scale"""
    x_sections = eels.get_x_sections()
    # print(energy_scale)
    energy_origin = energy_scale[0]
    energy_window = energy_scale[-1] - energy_origin
    selected_edges_unsorted = {}
    likely_edges = []
    selected_elements = []
    for element in range(1, 83):
        # print(element)
        element_z = str(eels.get_z(element))

        for key in x_sections[element_z]:
            if key in all_edges:
                onset = x_sections[element_z][key]['onset']
                if onset > energy_origin:
                    if onset - energy_origin < energy_window:
                        if element not in selected_edges_unsorted:
                            selected_edges_unsorted[element] = {}
                        # print(element, x_sections[element]['name'], key, x_sections[element][key]['onset'])
                        # text = f"\n {x_sections[element_z]['name']:2s}-{key}: " \
                        #       f"{x_sections[element_z][key]['onset']:8.1f} eV "
                        # print(text)

                        selected_edges_unsorted[element][key] = {}
                        selected_edges_unsorted[element][key]['onset'] = x_sections[element_z][key]['onset']

                        if key in major_edges:
                            selected_edges_unsorted[element][key]['intensity'] = 'major'
                            selected_elements.append(x_sections[element_z]['name'])
                        else:
                            selected_edges_unsorted[element][key]['intensity'] = 'minor'

        if element in selected_edges_unsorted:
            for key in selected_edges_unsorted[element]:
                if selected_edges_unsorted[element][key]['intensity'] == 'major':
                    likely_edges.append(x_sections[str(element)]['name'])  # = {'Z':element, 'symmetry': key}

    return likely_edges


def get_periodic_table_info():
    """Info for periodic table dialog"""
    pt_info = \
        {'H': {'PT_row': 0, 'PT_col': 0, 'Z': 0},
         'He': {'PT_row': 0, 'PT_col': 17, 'Z': 2}, 'Li': {'PT_row': 1, 'PT_col': 0, 'Z': 3},
         'Be': {'PT_row': 1, 'PT_col': 1, 'Z': 4}, 'B': {'PT_row': 1, 'PT_col': 12, 'Z': 5},
         'C': {'PT_row': 1, 'PT_col': 13, 'Z': 6}, 'N': {'PT_row': 1, 'PT_col': 14, 'Z': 7},
         'O': {'PT_row': 1, 'PT_col': 15, 'Z': 8}, 'F': {'PT_row': 1, 'PT_col': 16, 'Z': 9},
         'Ne': {'PT_row': 1, 'PT_col': 17, 'Z': 10}, 'Na': {'PT_row': 2, 'PT_col': 0, 'Z': 11},
         'Mg': {'PT_row': 2, 'PT_col': 1, 'Z': 12}, 'Al': {'PT_row': 2, 'PT_col': 12, 'Z': 13},
         'Si': {'PT_row': 2, 'PT_col': 13, 'Z': 14}, 'P': {'PT_row': 2, 'PT_col': 14, 'Z': 15},
         'S': {'PT_row': 2, 'PT_col': 15, 'Z': 16}, 'Cl': {'PT_row': 2, 'PT_col': 16, 'Z': 17},
         'Ar': {'PT_row': 2, 'PT_col': 17, 'Z': 18}, 'K': {'PT_row': 3, 'PT_col': 0, 'Z': 19},
         'Ca': {'PT_row': 3, 'PT_col': 1, 'Z': 20}, 'Sc': {'PT_row': 3, 'PT_col': 2, 'Z': 21},
         'Ti': {'PT_row': 3, 'PT_col': 3, 'Z': 22}, 'V ': {'PT_row': 3, 'PT_col': 4, 'Z': 23},
         'Cr': {'PT_row': 3, 'PT_col': 5, 'Z': 24}, 'Mn': {'PT_row': 3, 'PT_col': 6, 'Z': 25},
         'Fe': {'PT_row': 3, 'PT_col': 7, 'Z': 26}, 'Co': {'PT_row': 3, 'PT_col': 8, 'Z': 27},
         'Ni': {'PT_row': 3, 'PT_col': 9, 'Z': 28}, 'Cu': {'PT_row': 3, 'PT_col': 10, 'Z': 29},
         'Zn': {'PT_row': 3, 'PT_col': 11, 'Z': 30}, 'Ga': {'PT_row': 3, 'PT_col': 12, 'Z': 31},
         'Ge': {'PT_row': 3, 'PT_col': 13, 'Z': 32}, 'As': {'PT_row': 3, 'PT_col': 14, 'Z': 33},
         'Se': {'PT_row': 3, 'PT_col': 15, 'Z': 34}, 'Br': {'PT_row': 3, 'PT_col': 16, 'Z': 35},
         'Kr': {'PT_row': 3, 'PT_col': 17, 'Z': 36}, 'Rb': {'PT_row': 4, 'PT_col': 0, 'Z': 37},
         'Sr': {'PT_row': 4, 'PT_col': 1, 'Z': 38}, 'Y': {'PT_row': 4, 'PT_col': 2, 'Z': 39},
         'Zr': {'PT_row': 4, 'PT_col': 3, 'Z': 40}, 'Nb': {'PT_row': 4, 'PT_col': 4, 'Z': 41},
         'Mo': {'PT_row': 4, 'PT_col': 5, 'Z': 42}, 'Tc': {'PT_row': 4, 'PT_col': 6, 'Z': 43},
         'Ru': {'PT_row': 4, 'PT_col': 7, 'Z': 44}, 'Rh': {'PT_row': 4, 'PT_col': 8, 'Z': 45},
         'Pd': {'PT_row': 4, 'PT_col': 9, 'Z': 46}, 'Ag': {'PT_row': 4, 'PT_col': 10, 'Z': 47},
         'Cd': {'PT_row': 4, 'PT_col': 11, 'Z': 48}, 'In': {'PT_row': 4, 'PT_col': 12, 'Z': 49},
         'Sn': {'PT_row': 4, 'PT_col': 13, 'Z': 50}, 'Sb': {'PT_row': 4, 'PT_col': 14, 'Z': 51},
         'Te': {'PT_row': 4, 'PT_col': 15, 'Z': 52}, 'I': {'PT_row': 4, 'PT_col': 16, 'Z': 53},
         'Xe': {'PT_row': 4, 'PT_col': 17, 'Z': 54}, 'Cs': {'PT_row': 5, 'PT_col': 0, 'Z': 55},
         'Ba': {'PT_row': 5, 'PT_col': 1, 'Z': 56}, 'Hf': {'PT_row': 5, 'PT_col': 3, 'Z': 72},
         'Ta': {'PT_row': 5, 'PT_col': 4, 'Z': 73}, 'W': {'PT_row': 5, 'PT_col': 5, 'Z': 74},
         'Re': {'PT_row': 5, 'PT_col': 6, 'Z': 75}, 'Os': {'PT_row': 5, 'PT_col': 7, 'Z': 76},
         'Ir': {'PT_row': 5, 'PT_col': 8, 'Z': 77}, 'Pt': {'PT_row': 5, 'PT_col': 9, 'Z': 78},
         'Au': {'PT_row': 5, 'PT_col': 10, 'Z': 79}, 'Hg': {'PT_row': 5, 'PT_col': 11, 'Z': 80},
         'Pb': {'PT_row': 5, 'PT_col': 13, 'Z': 82}, 'Bi': {'PT_row': 5, 'PT_col': 14, 'Z': 0},
         'Po': {'PT_row': 5, 'PT_col': 15, 'Z': 0}, 'At': {'PT_row': 5, 'PT_col': 16, 'Z': 0},
         'Rn': {'PT_row': 5, 'PT_col': 17, 'Z': 0}, 'Fr': {'PT_row': 6, 'PT_col': 0, 'Z': 0},
         'Ra': {'PT_row': 6, 'PT_col': 1, 'Z': 0}, 'Rf': {'PT_row': 6, 'PT_col': 3, 'Z': 0},
         'Db': {'PT_row': 6, 'PT_col': 4, 'Z': 0}, 'Sg': {'PT_row': 6, 'PT_col': 5, 'Z': 0},
         'Bh': {'PT_row': 6, 'PT_col': 6, 'Z': 0}, 'Hs': {'PT_row': 6, 'PT_col': 7, 'Z': 0},
         'Mt': {'PT_row': 6, 'PT_col': 8, 'Z': 0}, 'Ds': {'PT_row': 6, 'PT_col': 9, 'Z': 0},
         'Rg': {'PT_row': 6, 'PT_col': 10, 'Z': 0}, 'La': {'PT_row': 8, 'PT_col': 3, 'Z': 57},
         'Ce': {'PT_row': 8, 'PT_col': 4, 'Z': 58}, 'Pr': {'PT_row': 8, 'PT_col': 5, 'Z': 59},
         'Nd': {'PT_row': 8, 'PT_col': 6, 'Z': 60}, 'Pm': {'PT_row': 8, 'PT_col': 7, 'Z': 61},
         'Sm': {'PT_row': 8, 'PT_col': 8, 'Z': 62}, 'Eu': {'PT_row': 8, 'PT_col': 9, 'Z': 63},
         'Gd': {'PT_row': 8, 'PT_col': 10, 'Z': 64}, 'Tb': {'PT_row': 8, 'PT_col': 11, 'Z': 65},
         'Dy': {'PT_row': 8, 'PT_col': 12, 'Z': 66}, 'Ho': {'PT_row': 8, 'PT_col': 13, 'Z': 67},
         'Er': {'PT_row': 8, 'PT_col': 14, 'Z': 68}, 'Tm': {'PT_row': 8, 'PT_col': 15, 'Z': 69},
         'Yb': {'PT_row': 8, 'PT_col': 16, 'Z': 70}, 'Lu': {'PT_row': 8, 'PT_col': 17, 'Z': 71},
         'Ac': {'PT_row': 9, 'PT_col': 3, 'Z': 0}, 'Th': {'PT_row': 9, 'PT_col': 4, 'Z': 0},
         'Pa': {'PT_row': 9, 'PT_col': 5, 'Z': 0}, 'U': {'PT_row': 9, 'PT_col': 6, 'Z': 0},
         'Np': {'PT_row': 9, 'PT_col': 7, 'Z': 0}, 'Pu': {'PT_row': 9, 'PT_col': 8, 'Z': 0},
         'Am': {'PT_row': 9, 'PT_col': 9, 'Z': 0}, 'Cm': {'PT_row': 9, 'PT_col': 10, 'Z': 0},
         'Bk': {'PT_row': 9, 'PT_col': 11, 'Z': 0}, 'Cf': {'PT_row': 9, 'PT_col': 12, 'Z': 0},
         'Es': {'PT_row': 9, 'PT_col': 13, 'Z': 0}, 'Fm': {'PT_row': 9, 'PT_col': 14, 'Z': 0},
         'Md': {'PT_row': 9, 'PT_col': 15, 'Z': 0}, 'No': {'PT_row': 9, 'PT_col': 16, 'Z': 0},
         'Lr': {'PT_row': 9, 'PT_col': 17, 'Z': 0},
         '*': {'PT_row': 5, 'PT_col': 2, 'PT_col2': 8, 'PT_row2': 2, 'Z': 0},
         '**': {'PT_row': 6, 'PT_col': 2, 'PT_col2': 9, 'PT_row2': 2, 'Z': 0}}

    return pt_info


class InteractiveSpectrumImage(object):
    """Interactive spectrum imaging plot

    Attributes:
    -----------
    dictionary with a minimum of the following keys:
        ['image']: displayed image
        ['data']:  data cube
        ['intensity_scale_ppm']: intensity scale
        ['ylabel']: intensity label
        ['spectra'] dictionary which contains dictionaries for each spectrum style ['1-2']:
            ['spectrum'] = tags['cube'][y,x,:]
            ['spectra'][f'{x}-{y}']['energy_scale'] = tags['energy_scale']
            ['intensity_scale'] = 1/tags['cube'][y,x,:].sum()*1e6

    Please note the possibility to load any image for the selection of the spectrum
    Also there is the possibility to display the survey image.

    For analysis we have the following options:
        'fix_energy': set zero-loss peak maximum to zero !! Low loss spectra only!!
        'fit_zero_loss': fit zero-loss peak with model function !! Low loss spectra only!!
        'fit_low_loss': fit low-loss spectrum with model peaks !! Low loss spectra only!!


        'fit_composition': fit core-loss spectrum with background and cross sections!! Core loss spectra only!!
        'fit_ELNES': fit core-loss edge with model peaks  !! Core loss spectra only!!
    """

    def __init__(self, data_source, horizontal=True):

        box_layout = widgets.Layout(display='flex',
                                    flex_flow='row',
                                    align_items='stretch',
                                    width='100%')

        words = ['fix_energy', 'fit_zero_loss', 'fit_low_loss', 'fit_composition', 'fit_ELNES']

        self.buttons = [widgets.ToggleButton(value=False, description=word, disabled=False) for word in words]
        box = widgets.Box(children=self.buttons, layout=box_layout)
        display(box)

        # MAKE Dictionary

        if isinstance(data_source, dict):
            self.tags = data_source
        elif isinstance(data_source, h5py.Group):
            self.tags = self.set_tags(data_source)
        else:
            print('Data source must be a dictionary or channel')
            return

        # Button(description='edge_quantification')
        for button in self.buttons:
            button.observe(self.on_button_clicked, 'value')  # on_click(self.on_button_clicked)

        self.figure = plt.figure()
        self.horizontal = horizontal
        self.x = 0
        self.y = 0

        self.extent = [0, self.tags['cube'].shape[1], self.tags['cube'].shape[0], 0]
        self.rectangle = [0, self.tags['cube'].shape[1], 0, self.tags['cube'].shape[0]]
        self.scaleX = 1.0
        self.scaleY = 1.0
        self.analysis = []
        self.plot_legend = False
        if 'ylabel' not in self.tags:
            self.tags['ylabel'] = 'intensity [a.u.]'
        self.SI = False

        if horizontal:
            self.ax1 = plt.subplot(1, 2, 1)
            self.ax2 = plt.subplot(1, 2, 2)
        else:
            self.ax1 = plt.subplot(2, 1, 1)
            self.ax2 = plt.subplot(2, 1, 2)

        self.cube = self.tags['cube']
        self.image = self.tags['cube'].sum(axis=2)

        self.ax1.imshow(self.image, extent=self.extent)
        if horizontal:
            self.ax1.set_xlabel('distance [pixels]')
        else:
            self.ax1.set_ylabel('distance [pixels]')
        self.ax1.set_aspect('equal')

        self.rect = patches.Rectangle((0, 0), 1, 1, linewidth=1, edgecolor='r', facecolor='red', alpha=0.2)
        self.ax1.add_patch(self.rect)
        self.intensity_scale = self.tags['spectra'][f'{self.x}-{self.y}']['intensity_scale']
        self.spectrum = self.tags['spectra'][f'{self.x}-{self.y}']['spectrum'] * self.intensity_scale
        self.energy_scale = self.tags['spectra'][f'{self.x}-{self.y}']['energy_scale']

        self.ax2.plot(self.energy_scale, self.spectrum)
        self.ax2.set_title(f' spectrum {self.x},{self.y} ')
        self.ax2.set_xlabel('energy loss [eV]')
        self.ax2.set_ylabel(self.tags['ylabel'])
        self.cid = self.figure.canvas.mpl_connect('button_press_event', self.onclick)

        plt.tight_layout()

    def on_button_clicked(self, b):
        # print(b['owner'].description)
        selection = b['owner'].description
        if b['new']:
            if selection == 'fit_composition':
                if 'region_tags' in self.tags and 'edges_present' in self.tags \
                        and 'acceleration_voltage' in self.tags \
                        and 'collection_angle' in self.tags:
                    pass
                else:
                    self.buttons[3].value = False
                    return
            elif selection in ['fix_energy', 'fit_zero_loss']:
                if self.energy_scale[0] > 0:
                    button_index = ['fix_energy', 'fit_zero_loss'].index(selection)
                    self.buttons[button_index].value = False
                    return
            self.analysis.append(selection)
            self.update()
        else:

            if selection in self.analysis:
                self.analysis.remove(selection)

    def do_all(self, selection=None, verbose=True):
        x = self.x
        y = self.y
        if selection is None:
            selection = self.analysis
        for self.x in range(self.cube.shape[1]):
            if verbose:
                print(f' row: {self.x}')
            for self.y in range(self.cube.shape[0]):

                if 'fit_zero_loss' in selection:
                    title = self.fit_zero_loss(plot_this=False)

                elif 'fix_energy' in selection:
                    self.ax2.set_title('bn')
                    title = self.fix_energy()

                elif 'fit_composition' in selection:
                    title = self.fit_quantification(plot_this=False)

        self.x = x
        self.y = y

    def onclick(self, event):
        x = int(event.xdata)
        y = int(event.ydata)

        # print(x,y)
        if self.rectangle[0] <= x < self.rectangle[0] + self.rectangle[1]:
            if self.rectangle[2] <= y < self.rectangle[2] + self.rectangle[3]:
                self.x = int((x - self.rectangle[0]) / self.rectangle[1] * self.cube.shape[1])
                self.y = int((y - self.rectangle[2]) / self.rectangle[3] * self.cube.shape[0])
            else:
                return
        else:
            return

        if event.inaxes in [self.ax1]:
            x = (self.x * self.rectangle[1] / self.cube.shape[1] + self.rectangle[0])
            y = (self.y * self.rectangle[3] / self.cube.shape[0] + self.rectangle[2])

            self.rect.set_xy([x, y])
            self.update()

    def update(self):
        xlim = self.ax2.get_xlim()
        ylim = self.ax2.get_ylim()
        self.ax2.clear()
        self.intensity_scale = self.tags['spectra'][f'{self.x}-{self.y}']['intensity_scale']
        self.spectrum = self.tags['spectra'][f'{self.x}-{self.y}']['spectrum'] * self.intensity_scale
        self.energy_scale = self.tags['spectra'][f'{self.x}-{self.y}']['energy_scale']

        if 'fit_zero_loss' in self.analysis:
            title = self.fit_zero_loss()
            self.ax2.set_title(title)
        elif 'fix_energy' in self.analysis:
            self.ax2.set_title('bn')
            title = self.fix_energy()
            self.ax2.set_title(title)

        elif 'fit_composition' in self.analysis:
            title = self.fit_quantification()
            self.ax2.set_title(title)

        else:
            self.ax2.set_title(f' spectrum {self.x},{self.y} ')
        self.ax2.plot(self.energy_scale, self.spectrum, color='#1f77b4', label='experiment')

        if self.plot_legend:
            self.ax2.legend(shadow=True)
        self.ax2.set_xlim(xlim)
        self.ax2.set_ylim(ylim)
        self.ax2.set_xlabel('energy loss [eV]')
        self.ax2.set_ylabel(self.tags['ylabel'])
        self.ax2.set_xlim(xlim)

        # self.ax2.draw()

    def set_tags(self, channel):
        # TODO: change to sidpy dataset tags = ft.h5_get_dictionary(channel)
        tags = {}
        if tags['data_type'] == 'spectrum_image':
            tags['image'] = tags['data']
            tags['data'] = tags['cube'][0, 0, :]
            if 'intentsity_scale_ppm' not in channel:
                channel['intentsity_scale_ppm'] = 1

            tags['ylabel'] = 'intensity [a.u.]'
            tags['spectra'] = {}
            for x in range(tags['spatial_size_y']):
                for y in range(tags['spatial_size_x']):
                    tags['spectra'][f'{x}-{y}'] = {}
                    tags['spectra'][f'{x}-{y}']['spectrum'] = tags['cube'][y, x, :]
                    tags['spectra'][f'{x}-{y}']['energy_scale'] = tags['energy_scale']
                    tags['spectra'][f'{x}-{y}']['intensity_scale'] = 1 / tags['cube'][y, x, :].sum() * 1e6
            tags['ylabel'] = 'inel. scat. int.  [ppm]'

        return tags

    def fix_energy(self):

        energy_scale = self.tags['spectra'][f'{self.x}-{self.y}']['energy_scale']
        spectrum = self.tags['spectra'][f'{self.x}-{self.y}']['spectrum'] * self.intensity_scale
        fwhm, delta_e = eels.fix_energy_scale(spectrum, energy_scale)
        self.tags['spectra'][f'{self.x}-{self.y}']['delta_e'] = delta_e
        self.tags['spectra'][f'{self.x}-{self.y}']['fwhm'] = fwhm
        self.energy_scale = energy_scale - delta_e
        title = f'spectrum {self.x},{self.y} fwhm: {fwhm:.2f}, dE: {delta_e:.3f}'
        return title

    def fit_zero_loss(self, plot_this=True):

        energy_scale = self.tags['spectra'][f'{self.x}-{self.y}']['energy_scale']
        spectrum = self.tags['spectra'][f'{self.x}-{self.y}']['spectrum'] * self.intensity_scale
        if 'zero_loss_fit_width' not in self.tags:
            self.tags['zero_loss_fit_width'] = .5
        if self.tags['zero_loss_fit_width'] / (energy_scale[1] - energy_scale[0]) < 6:
            self.tags['zero_loss_fit_width'] = (energy_scale[1] - energy_scale[0]) * 6
        fwhm, delta_e = eels.fix_energy_scale(spectrum, energy_scale)
        energy_scale = energy_scale - delta_e
        z_oss, p_zl = eels.resolution_function(energy_scale, spectrum, self.tags['zero_loss_fit_width'])
        fwhm2, delta_e2 = eels.fix_energy_scale(z_oss, energy_scale)

        self.tags['spectra'][f'{self.x}-{self.y}']['resolution_function'] = z_oss
        self.tags['spectra'][f'{self.x}-{self.y}']['p_zl'] = p_zl
        self.tags['spectra'][f'{self.x}-{self.y}']['delta_e'] = delta_e
        self.tags['spectra'][f'{self.x}-{self.y}']['fwhm_resolution'] = fwhm2
        self.tags['spectra'][f'{self.x}-{self.y}']['fwhm'] = fwhm

        if plot_this:
            self.ax2.plot(energy_scale, z_oss, label='resolution function', color='black')
            self.ax2.plot(energy_scale, self.spectrum - z_oss, label='difference', color='orange')
            self.ax2.axhline(linewidth=0.5, color='black')
        self.energy_scale = energy_scale
        title = f'spectrum {self.x},{self.y} fwhm: {fwhm:.2f}'  # ', dE: {delta_e2:.5e}'
        return title

    def fit_quantification(self, plot_this=True):
        energy_scale = self.tags['spectra'][f'{self.x}-{self.y}']['energy_scale']
        spectrum = self.tags['spectra'][f'{self.x}-{self.y}']['spectrum'] * self.intensity_scale
        edges = eels.make_edges(self.tags['edges_present'], energy_scale, self.tags['acceleration_voltage'],
                                self.tags['collection_angle'])
        edges = eels.fit_edges(spectrum, self.tags['spectra'][f'{self.x}-{self.y}']['energy_scale'],
                               self.tags['region_tags'], edges)
        self.tags['spectra'][f'{self.x}-{self.y}']['edges'] = edges.copy()
        if plot_this:
            self.ax2.plot(energy_scale, edges['model']['spectrum'], label='model')
            self.ax2.plot(energy_scale, self.spectrum - edges['model']['spectrum'], label='difference')
            self.ax2.axhline(linewidth=0.5, color='black')
        else:
            self.tags['spectra'][f'{self.x}-{self.y}']['do_all'] = 'done'
        title = f'spectrum {self.x},{self.y} '

        for key in edges:
            if key.isdigit():
                title = title + f"{edges[key]['element']}:  {edges[key]['areal_density']:.2e};  "

        return title

    def set_legend(self, set_legend):
        self.plot_legend = set_legend

    def get_xy(self):
        return [self.x, self.y]

    def get_current_spectrum(self):
        return self.cube[self.y, self.x, :]

    def set_z_contrast_image(self, z_channel=None):
        if z_channel is not None:
            self.tags['Z_contrast_channel'] = z_channel
        if 'Z_contrast_channel' not in self.tags:
            print('add Z contrast channel group to dictionary first!')
            return
        # get dictionary from current channel in pyUSID file
        z_tags = {}   # TODO change to sidpy dataset ft.h5_get_dictionary(z_channel)
        extent = [self.rectangle[0], self.rectangle[0] + self.rectangle[1],
                  self.rectangle[2] + self.rectangle[3], self.rectangle[2]]
        self.ax1.imshow(z_tags['data'], extent=extent, cmap='gray')

    def overlay_z_contrast_image(self, z_channel=None):

        if self.SI:
            if z_channel is not None:
                self.tags['Z_contrast_channel'] = z_channel
            if 'Z_contrast_channel' not in self.tags:
                print('add survey channel group to dictionary first!')
                return

            z_tags = {}  # TODO: change to sidpy ft.h5_get_dictionary(self.tags['Z_contrast_channel'])

            xlim = self.ax1.get_xlim()
            ylim = self.ax1.get_ylim()
            extent = [self.rectangle[0], self.rectangle[0] + self.rectangle[1],
                      self.rectangle[2] + self.rectangle[3], self.rectangle[2]]
            self.ax1.imshow(z_tags['data'], extent=extent, cmap='viridis', alpha=0.5)
            self.ax1.set_ylim(ylim)
            self.ax1.set_xlim(xlim)

    def overlay_data(self, data=None):

        if self.SI:
            if data is None:
                data = self.cube.sum(axis=2)

            xlim = self.ax1.get_xlim()
            ylim = self.ax1.get_ylim()
            extent = [self.rectangle[0], self.rectangle[0] + self.rectangle[1],
                      self.rectangle[2] + self.rectangle[3], self.rectangle[2]]
            self.ax1.imshow(data, extent=extent, alpha=0.7, cmap='viridis')
            self.ax1.set_ylim(ylim)
            self.ax1.set_xlim(xlim)

    def set_survey_image(self, si_channel=None):

        # get dictionary from current channel in pyUSID file
        if si_channel is not None:
            self.tags['survey_channel'] = si_channel
        if 'survey_channel' not in self.tags:
            print('add survey channel group to dictionary first!')
            return
        si_channel = self.tags['survey_channel']
        si_tags = {}  # TODO: change to sidpy  ft.h5_get_dictionary(si_channel)
        tags2 = dict(si_channel.attrs)

        self.ax1.set_aspect('equal')
        self.scaleX = si_channel['spatial_scale_x'][()]
        self.scaleY = si_channel['spatial_scale_y'][()]

        self.ax1.imshow(si_tags['data'], extent=si_tags['extent'], cmap='gray')
        if self.horizontal:
            self.ax1.set_xlabel('distance [nm]')
        else:
            self.ax1.set_ylabel('distance [nm]')

        annotation_done = []
        for key in tags2:
            if 'annotations' in key:
                annotation_number = key[12]
                if annotation_number not in annotation_done:
                    annotation_done.append(annotation_number)

                    if tags2['annotations_' + annotation_number + '_type'] == 'text':
                        x = tags2['annotations_' + annotation_number + '_x']
                        y = tags2['annotations_' + annotation_number + '_y']
                        text = tags2['annotations_' + annotation_number + '_text']
                        self.ax1.text(x, y, text, color='r')

                    elif tags2['annotations_' + annotation_number + '_type'] == 'circle':
                        radius = 20 * self.scaleX  # tags['annotations'][key]['radius']
                        xy = tags2['annotations_' + annotation_number + '_position']
                        circle = patches.Circle(xy, radius, color='r', fill=False)
                        self.ax1.add_artist(circle)

                    elif tags2['annotations_' + annotation_number + '_type'] == 'spectrum image':
                        width = tags2['annotations_' + annotation_number + '_width']
                        height = tags2['annotations_' + annotation_number + '_height']
                        position = tags2['annotations_' + annotation_number + '_position']
                        rectangle = patches.Rectangle(position, width, height, color='r', fill=False)
                        self.rectangle = [position[0], width, position[1], height]
                        self.ax1.add_artist(rectangle)
                        self.ax1.text(position[0], position[1], 'Spectrum Image', color='r')
                        self.rect.set_width(width / self.cube.shape[1])
                        self.rect.set_height(height / self.cube.shape[0])
        self.SI = True


class ElementalEdges(object):
    """ Adds ionization edges of element z to plot with axis ax

        There is an optional parameter maximum_chemical_shift which allows to change
        the energy range in which the edges are searched.

        available functions:
        - update(): updates the drawing of ionization edges
        - set_edge(Z) : changes atomic number and updates everything accordingly
        - disconnect: makes everything invisible and stops drawing
        - reconnect: undo of disconnect

        usage:
        >> fig, ax = plt.subplots()
        >> ax.plot(energy_scale, spectrum)
        >> Z= 42
        >> cursor = ElementalEdges(ax, Z)


        see Chapter4 'CH4-Working_with_X-Sections' notebook
    """

    def __init__(self, ax, z):
        self.ax = ax
        self.labels = None
        self.lines = None
        self.Z = eels.get_z(z)
        self.color = 'black'
        self.x_sections = eels.get_x_sections()
        self.cid = ax.figure.canvas.mpl_connect('draw_event', self.onresize)
        # self.update() is not necessary because of a drawing event is issued

    def set_edge(self, z):
        self.Z = eels.get_z(z)
        if self.cid is None:
            self.cid = self.ax.figure.canvas.mpl_connect('draw_event', self.onresize)
        self.update()

    def onresize(self, event):
        self.update()

    def update(self):
        if self.labels is not None:
            for label in self.labels:
                label.remove()
        if self.lines is not None:
            for line in self.lines:
                line.remove()
        self.labels = []
        self.lines = []
        x_min, x_max = self.ax.get_xlim()
        y_min, y_max = self.ax.get_ylim()

        element = str(self.Z)
        x_sections = self.x_sections
        for key in all_edges:
            if key in x_sections[element] and 'onset' in x_sections[element][key]:
                x = x_sections[element][key]['onset']
                if x_min < x < x_max:
                    if key in first_close_edges:
                        label2 = self.ax.text(x, y_max, f"{x_sections[element]['name']}-{key}",
                                              verticalalignment='top', rotation=0, color=self.color)
                    else:
                        label2 = self.ax.text(x, y_max, f"\n{x_sections[element]['name']}-{key}",
                                              verticalalignment='top', color=self.color)
                    line2 = self.ax.axvline(x, ymin=0, ymax=1, color=self.color)

                    self.labels.append(label2)
                    self.lines.append(line2)

    def reconnect(self):
        self.cid = self.ax.figure.canvas.mpl_connect('draw_event', self.onresize)
        self.update()

    def disconnect(self):
        if self.labels is not None:
            for label in self.labels:
                label.remove()
        if self.lines is not None:
            for line in self.lines:
                line.remove()
        self.labels = None
        self.lines = None
        self.ax.figure.canvas.mpl_disconnect(self.cid)


class EdgesAtCursor(object):
    """
        Adds a Cursor to a plot, which plots all major (possible) ionization edges at
        the cursor location if left (right) mouse button is clicked.

        Attributes
        ----------
        ax: matplotlib axis
        x: numpy array
            energy_scale of spectrum
        y: numpy array
            intensities of spectrum
        maximal_chemical_shift: float
            optional parameter maximum_chemical_shift which allows to change the energy range in which the edges
            are searched.

        Example
        -------
        fig, ax = plt.subplots()
        ax.plot(energy_scale, spectrum)
        cursor = EdgesAtCursor(ax, energy_scale, spectrum)

        see Chapter4 'CH4-Working_with_X-Sections' notebook

    """

    def __init__(self, ax, x, y, maximal_chemical_shift=5):
        self.ax = ax
        self.ly = ax.axvline(x[0], color='k', alpha=0.2)  # the vert line
        self.marker, = ax.plot(x[0], y[0], marker="o", color="crimson", zorder=3)
        self.x = x
        self.y = y
        self.txt = ax.text(0.7, 0.9, '', verticalalignment='bottom')
        self.select = 0
        self.label = None
        self.line = None
        self.cid = ax.figure.canvas.mpl_connect('button_press_event', self.click)
        self.mouse_cid = ax.figure.canvas.mpl_connect('motion_notify_event', self.mouse_move)
        self.maximal_chemical_shift = maximal_chemical_shift

    def click(self, event):

        # print('click', event)
        if not event.inaxes:
            return
        x, y = event.xdata, event.ydata

        indx = np.searchsorted(self.x, [x])[0]
        x = self.x[indx]
        y = self.y[indx]
        self.select = x

        y_min, y_max = self.ax.get_ylim()

        if self.label is not None:
            self.label.remove()
            self.line.remove()
        if event.button == 1:
            self.label = self.ax.text(x, y_max, eels.find_major_edges(event.xdata, self.maximal_chemical_shift),
                                      verticalalignment='top')
            self.line, = self.ax.plot([x, x], [y_min, y_max], color='black')
        if event.button == 3:
            self.line, = self.ax.plot([x, x], [y_min, y_max], color='black')
            self.label = self.ax.text(x, y_max, eels.find_all_edges(event.xdata, self.maximal_chemical_shift),
                                      verticalalignment='top')
        self.ax.set_ylim(y_min, y_max)

    def mouse_move(self, event):
        if not event.inaxes:
            return

        x, y = event.xdata, event.ydata
        indx = np.searchsorted(self.x, [x])[0]
        x = self.x[indx]
        y = self.y[indx]
        self.select = x
        self.ly.set_xdata(x)
        self.marker.set_data([x], [y])
        self.txt.set_text(f'\n x={x:1.2f}, y={y:1.2g}\n')

        # self.ax.text(x, y*2,find_major_edges(x))
        self.txt.set_position((x, y))
        self.ax.figure.canvas.draw_idle()

    def del_edges(self):
        if self.label is not None:
            self.label.remove()
            self.line.remove()
            self.label = None

    def disconnect(self):
        self.ly.remove()
        self.marker.remove()
        self.txt.remove()

        self.ax.figure.canvas.mpl_disconnect(self.cid)
        self.ax.figure.canvas.mpl_disconnect(self.mouse_cid)



"""file_tools: All tools to load and save data
"""
##################################
#
# 2018 01 31 Included Nion Swift files to be opened
# major revision 2020 09 to include sidpy and pyNSID data formats
#
##################################

import numpy as np
import h5py
import os
import sys

# For structure files of various flavor for instance POSCAR
import ase.io


# =============================================
#   Include sidpy and other pyTEMlib Libraries                                      #
# =============================================
from .config_dir import config_path

# from .nsi_reader import NSIDReader
from SciFiReaders import DM3Reader
from SciFiReaders import NionReader
import pyNSID
import sidpy


from .sidpy_tools import *

# Open/Save File dialog
try:
    from PyQt5 import QtGui, QtWidgets, QtCore
    QT_available = True
except ImportError:
    QT_available = False

if "google.colab" in sys.modules:
    QT_available = False

if QT_available:
    from .qt_sidpy_tools import *

Dimension = sidpy.Dimension
nest_dict = sidpy.base.dict_utils.nest_dict

get_slope = sidpy.base.num_utils.get_slope
__version__ = '2021.3.1'

flatten_dict = sidpy.dict_utils.flatten_dict
nest_dict = sidpy.dict_utils.nest_dict


class FileWidget(object):
    """Widget to select directories or widgets from a list

    Works in google colab.
    The widgets converts the name of the nion file to the one in Nion's swift software,
    because it is otherwise incomprehensible

    Attributes
    ----------
    dir_name: str
        name of starting directory
    extension: list of str
        extensions of files to be listed  in widget

    Methods
    -------
    get_directory
    set_options
    get_file_name

    Example
    -------
    >>from google.colab import drive
    >>drive.mount("/content/drive")
    >>file_list = pyTEMlib.file_tools.FileWidget()
    next code cell:
    >>dataset = pyTEMlib.file_tools.open_file(file_list.file_name)

    """

    def __init__(self, dir_name=None, extension=['*']):
        self.save_path = False
        self.dir_dictionary = {}
        self.dir_list = ['.', '..']
        self.display_list = ['.', '..']

        if dir_name is None:
            dir_name = get_last_path()
            self.save_path = True
            self.get_directory(dir_name)
        elif os.path.isdir(dir_name):
            self.get_directory(dir_name)
            self.dir_name = dir_name
        else:
            self.dir_name = '.'
            self.get_directory(self.dir_name)

        self.dir_list = ['.']
        self.extensions = extension
        self.file_name = ''

        self.select_files = widgets.Select(
            options=self.dir_list,
            value=self.dir_list[0],
            description='Select file:',
            disabled=False,
            rows=10,
            layout=widgets.Layout(width='70%')
        )
        display(self.select_files)
        self.set_options()
        self.select_files.observe(self.get_file_name, names='value')

    def get_directory(self, directory=None):
        self.dir_name = directory
        self.dir_dictionary = {}
        self.dir_list = []
        self.dir_list = ['.', '..'] + os.listdir(directory)

    def set_options(self):
        self.dir_name = os.path.abspath(os.path.join(self.dir_name, self.dir_list[self.select_files.index]))
        dir_list = os.listdir(self.dir_name)

        file_list = []
        display_file_list = []
        directory_list = []

        for i in range(len(dir_list)):
            name = dir_list[i]
            full_name = os.path.join(self.dir_name, name)

            if os.path.isfile(full_name):
                size = os.path.getsize(full_name) * 2 ** -20
                basename, extension = os.path.splitext(name)
                if self.extensions[0] == 'hf5':
                    if extension in ['.hf5']:
                        file_list.append(dir_list[i])
                        display_file_list.append(f" {name}  - {size:.1f} MB")
                else:
                    file_list.append(dir_list[i])
                    if extension in ['.h5', '.ndata']:
                        reader = NionReader(full_name)
                        dataset_nion = reader.read()
                        display_file_list.append(f" {dataset_nion.title}{extension}  - {size:.1f} MB")
                    elif extension in ['.hf5']:
                        display_file_list.append(f" {name}  -- {size:.1f} MB")
                    else:
                        display_file_list.append(f' {name}  - {size:.1f} MB')
            else:
                directory_list.append(name)

        sort = np.argsort(directory_list)
        self.dir_list = ['.', '..']
        self.display_list = ['.', '..']
        for j in sort:
            self.display_list.append(f' * {directory_list[j]}')
            self.dir_list.append(directory_list[j])

        sort = np.argsort(display_file_list)

        for i, j in enumerate(sort):
            if '--' in dir_list[j]:
                self.display_list.append(f' {i:3} {display_file_list[j]}')
            else:
                self.display_list.append(f' {i:3}   {display_file_list[j]}')
            self.dir_list.append(file_list[j])

        self.dir_label = os.path.split(self.dir_name)[-1] + ':'
        self.select_files.options = self.display_list

    def get_file_name(self, b):

        if os.path.isdir(os.path.join(self.dir_name, self.dir_list[self.select_files.index])):
            self.set_options()

        elif os.path.isfile(os.path.join(self.dir_name, self.dir_list[self.select_files.index])):
            self.file_name = os.path.join(self.dir_name, self.dir_list[self.select_files.index])


####
#  General Open and Save Methods
####


def get_qt_app():
    """will start QT Application if not running yet and returns QApplication """

    # start qt event loop
    _instance = QtWidgets.QApplication.instance()
    if not _instance:
        # print('not_instance')
        _instance = QtWidgets.QApplication([])

    return _instance


def get_last_path():
    """Returns the path of the file last opened"""
    try:
        fp = open(config_path + '\\path.txt', 'r')
        path = fp.read()
        fp.close()
    except IOError:
        path = ''

    if len(path) < 2:
        path = '.'
    return path


def save_path(filename):
    """Save path of last opened file"""

    if len(filename) > 1:
        fp = open(config_path + '\\path.txt', 'w')
        path, fname = os.path.split(filename)
        fp.write(path)
        fp.close()
    else:
        path = '.'
    return path


def savefile_dialog(initial_file='*.hf5', file_types=None):
    """Opens a save dialog in QT and returns name of file. New now with initial file"""

    # Check whether QT is available
    if not QT_available:
        print('No QT dialog')
        return None
    else:
        if file_types is None:
            file_types = "All files (*)"
    try:
        get_qt_app()
    except BaseException:
        pass

    # Determine last path used
    path = get_last_path()

    filename = sidpy.io.interface_utils.savefile_dialog(initial_file, file_types=file_types, file_path=path)
    save_path(filename)

    if len(filename) > 3:
        h5_file_name = get_h5_filename(filename)
        return h5_file_name
    else:
        return ''


def openfile_dialog(file_types=None):  # , multiple_files=False):
    """Opens a File dialog which is used in open_file() function

    This function uses pyQt5.
    The app of the Gui has to be running for QT. Tkinter does not run on Macs at this point in time.
    In jupyter notebooks use %gui Qt early in the notebook.

    The file looks first for a path.txt file for the last directory you used.

    Parameters
    ----------
    file_types : string
        file type filter in the form of '*.hf5'


    Returns
    -------
    filename : string
        full filename with absolute path and extension as a string

    Example
    -------
    >> import file_tools as ft
    >> filename = ft.openfile_dialog()
    >> print(filename)

    """

    # determine file types by extension
    if file_types is None:
        file_types = 'TEM files (*.dm3 *.qf3 *.ndata *.h5 *.hf5);;pyNSID files (*.hf5);;QF files ( *.qf3);;' \
                     'DM files (*.dm3);;Nion files (*.ndata *.h5);;All files (*)'
    elif file_types == 'pyNSID':
        file_types = 'pyNSID files (*.hf5);;TEM files (*.dm3 *.qf3 *.ndata *.h5 *.hf5);;QF files ( *.qf3);;' \
                     'DM files (*.dm3);;Nion files (*.ndata *.h5);;All files (*)'

        # file_types = [("TEM files",["*.dm*","*.hf*","*.ndata" ]),("pyUSID files","*.hf5"),("DM files","*.dm*"),
        # ("Nion files",["*.h5","*.ndata"]),("all files","*.*")]

    # Determine last path used
    path = get_last_path()
    _ = get_qt_app()

    filename = sidpy.io.interface_utils.openfile_dialog(file_types=file_types, file_path=path)

    save_path(filename)
    return filename


def save_dataset(dataset, filename=None,  h5_group=None):
    """Saves a dataset to a file in pyNSID format
    Parameters
    ----------
    dataset: sidpy.Dataset
        the data
    filename: str
        name of file to be opened, if filename is None, a QT file dialog will try to open
    h5_group: hd5py.Group
        not used yet

    """
    if filename is None:
        if QT_available:
            get_qt_app()
            filename = savefile_dialog()
        else:
            raise TypeError('filename must be provided if QT is not installed')
    h5_filename = get_h5_filename(filename)
    h5_file = h5py.File(h5_filename, mode='a')
    path, file_name = os.path.split(filename)
    basename, _ = os.path.splitext(file_name)

    if h5_group is None:
        if 'Measurement_000' in h5_file:
            h5_group = sidpy.hdf.prov_utils.create_indexed_group(h5_group, 'Measurement_')
        else:
            h5_group = h5_file.create_group('Measurement_000')

        if 'Channel_000' in h5_group:
            h5_group = sidpy.hdf.prov_utils.create_indexed_group(h5_group, 'Channel_')
        else:
            h5_group = h5_group.create_group('Channel_000')

    elif isinstance(h5_group, str):
        if h5_group not in h5_file:
            h5_group = h5_file.create_group(h5_group)
        else:

            if h5_group[-1] == '/':
                h5_group = h5_group[:-1]

            channel = h5_group.split('/')[-1]
            h5_group = h5_group[:-len(channel)]
            h5_group = sidpy.hdf.prov_utils.create_indexed_group(h5_group, 'Channel_')
    else:
        raise ValueError('h5_group needs to be string or None')
    dataset.original_metadata['original_title'] = dataset.title
    dataset.title = basename
    h5_dataset = pyNSID.hdf_io.write_nsid_dataset(dataset, h5_group)
    dataset.h5_dataset = h5_dataset
    return h5_dataset


def open_file(filename=None,  h5_group=None):  # save_file=False,
    """Opens a file if the extension is .hf5, .ndata, .dm3 or .dm4

    If no filename is provided the QT open_file windows opens (if QT_available==True)
    Everything will be stored in a NSID style hf5 file.
    Subroutines used:
        - NSIDReader
        - nsid.write_
            - get_main_tags
            - get_additional tags

    Parameters
    ----------
    filename: str
        name of file to be opened, if filename is None, a QT file dialog will try to open
    h5_group: hd5py.Group
        not used yet #TODO: provide hook for usage of external chosen group

    Returns
    -------
    sidpy.Dataset
        sidpy dataset with location of hdf5 dataset as attribute

    """
    if filename is None:
        if QT_available:
            get_qt_app()
            filename = openfile_dialog()
        else:
            raise TypeError('filename must be provided if QT is not installed')
    else:
        if not isinstance(filename, str):
            raise TypeError('filename must be a non-empty string or None (to a QT open file dialog)')
        elif filename == '':
            raise TypeError('filename must be a non-empty string or None (to a QT open file dialog)')

    path, file_name = os.path.split(filename)
    basename, extension = os.path.splitext(file_name)

    if extension == '.hf5':

        reader = pyNSID.NSIDReader(filename)
        datasets = reader.read()
        if len(datasets) < 1:
            print('no hdf5 dataset found in file')
            return
        else:
            return datasets[0]

        """ should go to no dataset found
        if 'Raw_Data' in h5_group:
            dataset = read_old_h5group(h5_group)
            dataset.h5_dataset = h5_group['Raw_Data']
        """

    elif extension in ['.dm3', '.dm4', '.ndata', '.ndata1', '.h5']:

        # tags = open_file(filename)
        if extension in ['.dm3', '.dm4']:
            reader = DM3Reader(filename)
        else:   # extension in ['.ndata', '.h5']:
            reader = NionReader(filename)

        path, file_name = os.path.split(filename)
        basename, _ = os.path.splitext(file_name)
        dset = reader.read()

        if extension in ['.dm3', '.dm4']:
            dset.title = basename.strip().replace('-', '_')
        dset.filename = basename.strip().replace('-', '_')
        # dset.original_metadata = flatten_dict(dset.original_metadata)
        filename = os.path.join(path,  dset.title+extension)
        h5_filename = get_h5_filename(filename)
        h5_file = h5py.File(h5_filename, mode='a')

        if 'Measurement_000' in h5_file:
            print('could not write dataset to file, try saving it with ft.save()')
        else:
            if not isinstance(h5_group, h5py.Group):
                h5_group = h5_file.create_group('Measurement_000/Channel_000')
            # dset.axes = dset._axes
            # dset.attrs = {}
            h5_dataset = pyNSID.hdf_io.write_nsid_dataset(dset, h5_group)
            # dset.original_metadata = nest_dict(dset.original_metadata)

            dset.h5_dataset = h5_dataset
        return dset
    else:
        print('file type not handled yet.')
        return


def get_h5_filename(fname):
    """Determines file name of hdf5 file for newly converted data file"""

    path, filename = os.path.split(fname)
    basename, extension = os.path.splitext(filename)
    h5_file_name_original = os.path.join(path, basename + '.hf5')
    h5_file_name = h5_file_name_original

    if os.path.exists(os.path.abspath(h5_file_name_original)):
        count = 1
        h5_file_name = h5_file_name_original[:-4] + '-' + str(count) + '.hf5'
        while os.path.exists(os.path.abspath(h5_file_name)):
            count += 1
            h5_file_name = h5_file_name_original[:-4] + '-' + str(count) + '.hf5'

    if h5_file_name != h5_file_name_original:
        path, filename = os.path.split(h5_file_name)
        print('Cannot overwrite file. Using: ', filename)
    return str(h5_file_name)


def get_start_channel(h5_file):
    """ Legacy for get start channel"""

    DeprecationWarning('Depreciated: use function get_main_channel instead')
    return get_main_channel(h5_file)


def get_main_channel(h5_file):
    """Returns name of first channel group in hdf5-file"""

    current_channel = None
    if 'Measurement_000' in h5_file:
        if 'Measurement_000/Channel_000' in h5_file:
            current_channel = h5_file['Measurement_000/Channel_000']
    return current_channel


def h5_tree(input_object):
    """Just a wrapper for the sidpy function print_tree,

    so that sidpy does not have to be loaded in notebook

    """

    if isinstance(input_object, sidpy.Dataset):
        if not isinstance(input_object.h5_dataset, h5py.Dataset):
            raise ValueError('sidpy dataset does not have an associated h5py dataset')
        h5_file = input_object.h5_dataset.file
    elif isinstance(input_object, h5py.Dataset):
        h5_file = input_object.file
    elif isinstance(input_object, (h5py.Group, h5py.File)):
        h5_file = input_object
    else:
        raise TypeError('should be a h5py.object or sidpy Dataset')
    sidpy.hdf_utils.print_tree(h5_file)


def log_results(h5_group, dataset=None, attributes=None):
    """Log Results in hdf5-file

    Saves either a sidpy.Dataset or dictionary in a hdf5-file.
    The group for the result will consist of 'Log_ and a running index.
    That group will be placed in h5_group.

    Parameters
    ----------
    h5_group: hd5py.Group
        groups where result group are to be stored
    dataset: sidpy.Dataset or None
        sidpy dataset to be stored
    attributes: dict
        dictionary containing results that are not based on a sidpy.Dataset

    Returns
    -------
    log_group: hd5py.Group
        group in hdf5 file with results.

    """
    if isinstance(h5_group, sidpy.Dataset):
        h5_group = h5_group.h5_dataset
        if not isinstance(h5_group, h5py.Dataset):
            raise TypeError('Use h5_dataset of sidpy.Dataset is not a valid h5py.Dataset')
        h5_group = h5_group.parent.parent

    if not isinstance(h5_group, h5py.Group):
        raise TypeError('Need a valid h5py.Group for logging results')

    if dataset is None:
        log_group = sidpy.hdf.prov_utils.create_indexed_group(h5_group, 'Log_')
    else:
        log_group = pyNSID.hdf_io.write_results(h5_group, dataset=dataset)
        if hasattr(dataset, 'meta_data'):
            if 'analysis' in dataset.meta_data:
                log_group['analysis'] = dataset.meta_data['analysis']
        dataset.h5_dataset = log_group[dataset.title.replace('-', '_')][dataset.title.replace('-', '_')]
    if attributes is not None:
        for key, item in attributes.items():
            if not isinstance(item, dict):
                log_group[key] = attributes[key]
            else:
                flat_dict = sidpy.hdf.hdf_utils.flatten_dict(attributes[key])
                log_group.create_group(key)
                sidpy.hdf.hdf_utils.write_simple_attrs(log_group[key], flat_dict)
    return log_group


def add_dataset(dataset, h5_group=None):
    """Write data to hdf5 file

    Parameters
    ----------
    dataset: sidpy.Dataset
        data to write to file
    h5_group: None, sidpy.Dataset, h5py.Group, h5py.Datset, h5py.File
        identifier to which group the data are added (if None the dataset must have a valid h5_dataset)

    Returns
    -------
    log_group: h5py.Dataset
        reference the dataset has been written to. (is also stored in h5_dataset attribute of sipy.Dataset)
    """

    if h5_group is None:
        if isinstance(dataset.h5_dataset, h5py.Dataset):
            h5_group = dataset.h5_dataset.parent.parent.parent
    if isinstance(h5_group, h5py.Dataset):
        h5_group = h5_group.parent.parent.parent
    elif isinstance(h5_group, sidpy.Dataset):
        h5_group = h5_group.h5_dataset.parent.parent.parent
    elif isinstance(h5_group, h5py.File):
        h5_group = h5_group['Measurement_000']

    if not isinstance(h5_group, h5py.Group):
        raise TypeError('Need a valid identifier for a hdf5 group to store data in')

    log_group = sidpy.hdf.prov_utils.create_indexed_group(h5_group, 'Channel_')
    h5_dataset = pyNSID.hdf_io.write_nsid_dataset(dataset, log_group)

    if hasattr(dataset, 'meta_data'):
        if 'analysis' in dataset.meta_data:
            log_group['analysis'] = dataset.meta_data['analysis']

    dataset.h5_dataset = h5_dataset
    return h5_dataset


###
# Crystal Structure Read and Write
###
def read_poscar(file_name=None):
    """
    Open file dialog to select a POSCAR file from VASP
    """
    if file_name is None:
        file_name = openfile_dialog('POSCAR (POSCAR*.txt);;All files (*)')

    # use ase package to read file
    base = os.path.basename(file_name)
    base_name = os.path.splitext(base)[0]
    crystal = ase.io.read(file_name, format='vasp', parallel=False)

    # make dictionary and plot structure (not essential for further notebook)
    tags = {'unit_cell': crystal.cell * 1e-1, 'elements': crystal.get_chemical_symbols(),
            'base': crystal.get_scaled_positions(), 'max_bond_length': 0.23, 'name': base_name}
    return tags


def read_cif(file_name=None, verbose=False):  # open file dialog to select cif file
    if file_name is None:
        file_name = openfile_dialog('cif (*.cif);;All files (*)')
    # use ase package to read file

    base = os.path.basename(file_name)
    base_name = os.path.splitext(base)[0]
    crystal = ase.io.read(file_name, format='cif', store_tags=True, parallel=False)

    # make dictionary and plot structure (not essential for further notebook)
    tags = {'unit_cell': crystal.cell * 1e-1, 'elements': crystal.get_chemical_symbols(),
            'base': crystal.get_scaled_positions(), 'max_bond_length': 0.23, 'crystal_name': base_name, 'reference': {},
            'info': {}}
    if verbose:
        print('Opened cif file for ', crystal.get_chemical_formula())

    for key in crystal.info:
        if 'citat' in key or 'pub' in key:
            tags['reference'][key] = crystal.info[key]
            if verbose:
                print(key, crystal.info[key])
        else:
            tags['info'][key] = crystal.info[key]

    return tags


def h5_add_crystal_structure(h5_file, crystal_tags):
    """Write crystal structure to NSID file"""

    structure_group = sidpy.hdf.prov_utils.create_indexed_group(h5_file, 'Structure_')

    structure_group['unit_cell'] = crystal_tags['unit_cell']
    structure_group['relative_positions'] = crystal_tags['base']
    structure_group['title'] = str(crystal_tags['crystal_name'])
    structure_group['_' + crystal_tags['crystal_name']] = str(crystal_tags['crystal_name'])
    structure_group['elements'] = np.array(crystal_tags['elements'], dtype='S')
    if 'zone_axis' in crystal_tags:
        structure_group['zone_axis'] = np.array(crystal_tags['zone_axis'], dtype=float)
    else:
        structure_group['zone_axis'] = np.array([1., 0., 0.], dtype=float)
    h5_file.flush()
    return structure_group


def h5_add_to_structure(structure_group, crystal_tags):
    """add dictionary as structure group"""

    for key in crystal_tags:
        if key in structure_group.keys():
            print(key, ' not written; use new name')
        else:
            structure_group[key] = crystal_tags[key]


def h5_get_crystal_structure(structure_group):
    """Read crystal structure  from NSID file"""

    crystal_tags = {'unit_cell': structure_group['unit_cell'][()],
                    'base': structure_group['relative_positions'][()],
                    'crystal_name': structure_group['title'][()]}
    if '2D' in structure_group:
        crystal_tags['2D'] = structure_group['2D'][()]
    elements = structure_group['elements'][()]
    crystal_tags['elements'] = []
    for e in elements:
        crystal_tags['elements'].append(e.astype(str, copy=False))

    if 'zone_axis' in structure_group:
        crystal_tags['zone_axis'] = structure_group['zone_axis'][()]
    return crystal_tags


###############################################
# Support old pyTEM file format
###############################################

def read_old_h5group(current_channel):
    """Make a  sidpy.Dataset from pyUSID style hdf5 group

    Parameters
    ----------
        current_channel: h5_group

    Returns
    -------
        sidpy.Dataset
    """

    dim_dir = []
    if 'nDim_Data' in current_channel:
        h5_dataset = current_channel['nDim_Data']
        reader = pyNSID.NSIDReader(h5_dataset.file.filename)
        dataset = reader.read(h5_dataset)
        dataset.h5_file = current_channel.file
        return dataset
    elif 'Raw_Data' in current_channel:
        if 'image_stack' in current_channel:
            sid_dataset = sidpy.Dataset.from_array(np.swapaxes(current_channel['image_stack'][()], 2, 0))
            dim_dir = ['SPATIAL', 'SPATIAL', 'TEMPORAL']
        elif 'data' in current_channel:
            sid_dataset = sidpy.Dataset.from_array(current_channel['data'][()])
            dim_dir = ['SPATIAL', 'SPATIAL']
        else:
            size_x = int(current_channel['spatial_size_x'][()])
            size_y = int(current_channel['spatial_size_y'][()])
            if 'spectral_size_x' in current_channel:
                size_s = int(current_channel['spectral_size_x'][()])
            else:
                size_s = 0
            data = np.reshape(current_channel['Raw_Data'][()], (size_x, size_y, size_s))
            sid_dataset = sidpy.Dataset.from_array(data)
            if size_x > 1:
                dim_dir.append('SPATIAL')
            if size_y > 1:
                dim_dir.append('SPATIAL')
            if size_s > 1:
                dim_dir.append('SPECTRAL')
        sid_dataset.h5_dataset = current_channel['Raw_Data']

    elif 'data' in current_channel:
        sid_dataset = sidpy.Dataset.from_array(current_channel['data'][()])
        dim_dir = ['SPATIAL', 'SPATIAL']
        sid_dataset.h5_dataset = current_channel['data']
    else:
        return

    if 'SPATIAL' in dim_dir:
        if 'SPECTRAL' in dim_dir:
            sid_dataset.data_type = sidpy.DataType.SPECTRAL_IMAGE
        elif 'TEMPORAL' in dim_dir:
            sid_dataset.data_type = sidpy.DataType.IMAGE_STACK
        else:
            sid_dataset.data_type = sidpy.DataType.IMAGE
    else:
        sid_dataset.data_type = sidpy.DataType.SPECTRUM

    sid_dataset.quantity = 'intensity'
    sid_dataset.units = 'counts'
    if 'analysis' in current_channel:
        sid_dataset.source = current_channel['analysis'][()]

    set_dimensions(sid_dataset, current_channel)

    return sid_dataset


def set_dimensions(dset, current_channel):
    """Attaches correct dimension from old pyTEMlib style.

    Parameters
    ----------
    dset: sidpy.Dataset
    current_channel: hdf5.Group
    """
    dim = 0
    if dset.data_type == sidpy.DataType.IMAGE_STACK:
        dset.set_dimension(dim, sidpy.Dimension(np.arange(dset.shape[dim]), name='frame',
                                                units='frame', quantity='stack',
                                                dimension_type='TEMPORAL'))
        dim += 1
    if 'IMAGE' in dset.data_type:

        if 'spatial_scale_x' in current_channel:
            scale_x = current_channel['spatial_scale_x'][()]
        else:
            scale_x = 1
        if 'spatial_units' in current_channel:
            units_x = current_channel['spatial_units'][()]
            if len(units_x) < 2:
                units_x = 'pixel'
        else:
            units_x = 'generic'
        if 'spatial_scale_y' in current_channel:
            scale_y = current_channel['spatial_scale_y'][()]
        else:
            scale_y = 0
        dset.set_dimension(dim, sidpy.Dimension('x', np.arange(dset.shape[dim])*scale_x,
                                                units=units_x, quantity='Length',
                                                dimension_type='SPATIAL'))
        dim += 1
        dset.set_dimension(dim, sidpy.Dimension('y', np.arange(dset.shape[dim])*scale_y,
                                                units=units_x, quantity='Length',
                                                dimension_type='SPATIAL'))
        dim += 1
    if dset.data_type in [sidpy.DataType.SPECTRUM, sidpy.DataType.SPECTRAL_IMAGE]:
        if 'spectral_scale_x' in current_channel:
            scale_s = current_channel['spectral_scale_x'][()]
        else:
            scale_s = 1.0
        if 'spectral_units_x' in current_channel:
            units_s = current_channel['spectral_units_x']
        else:
            units_s = 'eV'

        if 'spectral_offset_x' in current_channel:
            offset = current_channel['spectral_offset_x']
        else:
            offset = 0.0
        dset.set_dimension(dim, sidpy.Dimension(np.arange(dset.shape[dim]) * scale_s + offset,
                                                name='energy',
                                                units=units_s,
                                                quantity='energy_loss',
                                                dimension_type='SPECTRAL'))



import matplotlib.pyplot as plt

import matplotlib as mpl
import matplotlib.patches as patches
from mpl_toolkits.axes_grid1.anchored_artists import AnchoredSizeBar
from matplotlib.patches import Circle  # , Ellipse, Rectangle
from matplotlib.collections import PatchCollection
from matplotlib.lines import Line2D

from scipy.ndimage.interpolation import geometric_transform

import numpy as np

###################################
# Plot Reciprocal Unit Cell in 2D #
###################################


def plot_reciprocal_unit_cell_2D(tags):
    # Plot # unit cell in reciprocal space in 2D
    reciprocal_unit_cell = tags['reciprocal_unit_cell']

    # ignore y direction

    x = [reciprocal_unit_cell[0, 0], reciprocal_unit_cell[0, 0], reciprocal_unit_cell[1, 0], reciprocal_unit_cell[1, 0]]
    z = [reciprocal_unit_cell[0, 2], reciprocal_unit_cell[2, 2], reciprocal_unit_cell[2, 2], reciprocal_unit_cell[0, 2]]

    print(x, z)
    print(reciprocal_unit_cell)

    # Plot 2D
    fig = plt.figure()
    # ax = fig.add_subplot(111)
    ax = plt.gca()  # current axis

    ax.scatter(x, z, c='red', s=80)
    ax.add_patch(
        patches.Rectangle(
            (0, 0),  # (x,y)
            reciprocal_unit_cell[0, 0],  # width
            reciprocal_unit_cell[2, 2],  # height
            fill=False  # remove background
        )
    )
    ax.add_patch(
        patches.FancyArrow(0, 0, reciprocal_unit_cell[0, 0], 0, width=0.02,
                           color='black',
                           head_width=0.08,  # Default: 3 * width
                           head_length=0.1,  # Default: 1.5 * head_width
                           length_includes_head=True  # Default: False
                           )
    )
    ax.add_patch(
        patches.FancyArrow(0, 0, 0, reciprocal_unit_cell[2, 2], width=0.02,
                           color='black',
                           head_width=0.08,  # Default: 3 * width
                           head_length=0.1,  # Default: 1.5 * head_width
                           length_includes_head=True  # Default: False
                           )
    )

    plt.xlabel('x 1/nm')
    plt.ylabel('z 1/nm')
    ax.axis('equal')
    # plt.title('Unit Cell in Reciprocal Space of {0}'.format(tags['crystal']) )
    # texfig.savefig("recip_unit_cell")
    # fig.savefig('recip_unit_cell.jpg', dpi=90, bbox_inches='tight')
    plt.show()


#####################
# Plot SAED Pattern #
#####################
def plotSAED_parameter(tags, gray=False):
    tags['convergence_angle_nm-1'] = 0

    tags['background'] = 'white'  # 'white'  'grey'
    tags['color map'] = 'plasma'  # ,'cubehelix'#'Greys'#'plasma'
    tags['color reflections'] = 'intensity'

    if gray:
        tags['color map'] = 'gray'
        tags['background'] = '#303030'  # 'darkgray'
        tags['color reflections'] = 'intensity'
    tags['plot HOLZ'] = 0
    tags['plot HOLZ excess'] = 0
    tags['plot Kikuchi'] = 1
    tags['plot reflections'] = 1

    tags['color Kikuchi'] = 'green'

    tags['linewidth HOLZ'] = -1  # -1: linewidth according to intensity (structure factor F^2
    tags['linewidth Kikuchi'] = -1  # -1: linewidth according to intensity (structure factor F^2

    tags['label HOLZ'] = 0
    tags['label Kikuchi'] = 0
    tags['label reflections'] = 0

    tags['label color'] = 'white'
    tags['label size'] = 10

    tags['color Laue Zones'] = ['red', 'blue', 'green', 'blue', 'green']  # , 'green', 'red'] #for OLZ give a sequence
    tags['color zero'] = 'red'  # 'None' #'white'
    tags['color ring zero'] = 'None'  # 'Red' #'white' #, 'None'
    tags['width ring zero'] = 2

    # plotDiffPattern(tags,True)
    tags['plot rotation'] = 0.  # degree
    tags['plot shift x'] = -0.0
    tags['plot shift y'] = 0.0


########################
# Plot Kikuchi Pattern #
########################
def plotKikuchi(tags, grey=False):
    tags['background'] = 'black'  # 'white'  'grey'
    tags['color map'] = 'plasma'  # ,'cubehelix'#'Greys'#'plasma'
    tags['color reflections'] = 'intensity'

    tags['plot HOLZ'] = 0
    tags['plot HOLZ excess'] = 0
    tags['plot Kikuchi'] = 1
    tags['plot reflections'] = 1

    tags['label HOLZ'] = 0
    tags['label Kikuchi'] = 0
    tags['label reflections'] = 0

    tags['label color'] = 'white'
    tags['label size'] = 10

    tags['color Kikuchi'] = 'green'
    tags['linewidth HOLZ'] = -1  # -1: linewidth according to intensity (structure factor F^2
    tags['linewidth Kikuchi'] = -1  # -1: linewidth according to intensity (structure factor F^2

    tags['color Laue Zones'] = ['red', 'blue', 'green', 'blue', 'green']  # , 'green', 'red'] #for OLZ give a sequence
    tags['color zero'] = 'white'  # 'None' #'white'
    tags['color ring zero'] = 'None'  # 'Red' #'white' #, 'None'
    tags['width ring zero'] = 2

    # plotDiffPattern(tags,True)


########################
# Plot HOLZ Pattern #
########################

def plotHOLZ_parameter(tags, grey=False):
    tags['background'] = 'gray'  # 'white'  'grey'
    tags['color map'] = 'plasma'  # ,'cubehelix'#'Greys'#'plasma'
    tags['color reflections'] = 'intensity'

    tags['plot HOLZ'] = 1
    tags['plot HOLZ excess'] = 1
    tags['plot Kikuchi'] = 1
    tags['plot reflections'] = 1

    tags['label HOLZ'] = 0
    tags['label Kikuchi'] = 0
    tags['label reflections'] = 0

    tags['label color'] = 'white'
    tags['label size'] = 12

    tags['color Kikuchi'] = 'green'
    tags['linewidth HOLZ'] = 1  # -1: linewidth according to intensity (structure factor F^2
    tags['linewidth Kikuchi'] = -1  # -1: linewidth according to intensity (structure factor F^2

    tags['color Laue Zones'] = ['red', 'blue', 'lightblue', 'green', 'red']  # for OLZ give a sequence
    tags['color zero'] = 'None'  # 'white'
    tags['color ring zero'] = 'Red'  # 'Red' #'white' #, 'None'
    tags['width ring zero'] = 2

    tags['plot rotation'] = 0.  # degree
    tags['plot shift x'] = -0.0
    tags['plot shift y'] = 0.0

    # plotDiffPattern(holz,True)


########################
# Plot CBED Pattern #
########################

def plotCBED_parameter(tags):
    tags['background'] = 'black'  # 'white'  'grey'
    tags['color map'] = 'plasma'  # ,'cubehelix'#'Greys'#'plasma'
    tags['color reflections'] = 'intensity'

    tags['plot HOLZ'] = 1
    tags['plot HOLZ excess'] = 1
    tags['plot Kikuchi'] = 1
    tags['plot reflections'] = 1

    tags['label HOLZ'] = 0
    tags['label Kikuchi'] = 0
    tags['label reflections'] = 0

    tags['label color'] = 'white'
    tags['label size'] = 10

    tags['color Kikuchi'] = 'green'
    tags['linewidth HOLZ'] = -1  # -1: linewidth according to intensity (structure factor F^2
    tags['linewidth Kikuchi'] = -1  # -1: linewidth according to intensity (structure factor F^2

    tags['color reflections'] = 'intensity'

    tags['color Laue Zones'] = ['red', 'blue', 'green']  # , 'green', 'red'] #for OLZ give a sequence
    tags['color zero'] = 'white'  # 'None' #'white'
    tags['color ring zero'] = 'Red'  # 'Red' #'white' #, 'None'
    tags['width ring zero'] = 2

    # plotDiffPattern(tags,True)


########################
# Plot HOLZ Pattern #
########################


def circles(x, y, s, c='b', vmin=None, vmax=None, **kwargs):
    """
    Make a scatter plot of circles.
    Similar to plt.scatter, but the size of circles are in data scale.
    Parameters
    ----------
    x, y : scalar or array_like, shape (n, )
        Input data
    s : scalar or array_like, shape (n, )
        Radius of circles.
    c : color or sequence of color, optional, default : 'b'
        `c` can be a single color format string, or a sequence of color
        specifications of length `N`, or a sequence of `N` numbers to be
        mapped to colors using the `cmap` and `norm` specified via kwargs.
        Note that `c` should not be a single numeric RGB or RGBA sequence
        because that is indistinguishable from an array of values
        to be colormapped. (If you insist, use `color` instead.)
        `c` can be a 2-D array in which the rows are RGB or RGBA, however.
    vmin, vmax : scalar, optional, default: None
        `vmin` and `vmax` are used in conjunction with `norm` to normalize
        luminance data.  If either are `None`, the min and max of the
        color array is used.
    kwargs : `~matplotlib.collections.Collection` properties
        Eg. alpha, edgecolor(ec), facecolor(fc), linewidth(lw), linestyle(ls),
        norm, cmap, transform, etc.
    Returns
    -------
    paths : `~matplotlib.collections.PathCollection`
    Examples
    --------
    a = np.arange(11)
    circles(a, a, s=a*0.2, c=a, alpha=0.5, ec='none')
    plt.colorbar()
    License
    --------
    This code is under [The BSD 3-Clause License]
    (http://opensource.org/licenses/BSD-3-Clause)
    """

    if np.isscalar(c):
        kwargs.setdefault('color', c)
        c = None

    if 'fc' in kwargs:
        kwargs.setdefault('facecolor', kwargs.pop('fc'))
    if 'ec' in kwargs:
        kwargs.setdefault('edgecolor', kwargs.pop('ec'))
    if 'ls' in kwargs:
        kwargs.setdefault('linestyle', kwargs.pop('ls'))
    if 'lw' in kwargs:
        kwargs.setdefault('linewidth', kwargs.pop('lw'))
    # You can set `facecolor` with an array for each patch,
    # while you can only set `facecolors` with a value for all.

    zipped = np.broadcast(x, y, s)
    patches = [Circle((x_, y_), s_, picker=True)
               for x_, y_, s_ in zipped]
    collection = PatchCollection(patches, **kwargs)
    if c is not None:
        c = np.broadcast_to(c, zipped.shape).ravel()
        collection.set_array(c)
        collection.set_clim(vmin, vmax)

    ax = plt.gca()
    ax.add_collection(collection)
    ax.autoscale_view()
    plt.draw_if_interactive()
    if c is not None:
        plt.sci(collection)
    return collection


def topolar(img, order=1):
    """
    Transform img to its polar coordinate representation.

    order: int, default 1
        Specify the spline interpolation order.
        High orders may be slow for large images.
    """
    # max_radius is the length of the diagonal
    # from a corner to the mid-point of img.
    max_radius = 0.5 * np.linalg.norm(img.shape)

    def transform(coords):
        # Put coord[1] in the interval, [-pi, pi]
        theta = 2 * np.pi * coords[1] / (img.shape[1] - 1.)

        # Then map it to the interval [0, max_radius].
        # radius = float(img.shape[0]-coords[0]) / img.shape[0] * max_radius
        radius = max_radius * coords[0] / img.shape[0]

        i = 0.5 * img.shape[0] - radius * np.sin(theta)
        j = radius * np.cos(theta) + 0.5 * img.shape[1]
        return i, j

    polar = geometric_transform(img, transform, order=order)

    rads = max_radius * np.linspace(0, 1, img.shape[0])
    angs = np.linspace(0, 2 * np.pi, img.shape[1])

    return polar, (rads, angs)


def plotRingPattern(tags, grey=False):
    d = tags['Ring_Pattern']['allowed']['g norm']
    label = tags['Ring_Pattern']['allowed']['label']
    if 'label color' not in tags:
        tags['label color'] = 'navy'
    if 'profile color' not in tags:
        tags['profile color'] = 'navy'
    if 'ring color' not in tags:
        tags['ring color'] = 'red'
    if 'label size' not in tags:
        tags['label size'] = 10
    if 'profile height' not in tags:
        tags['profile height'] = 5
    if 'plot scalebar' not in tags:
        tags['plot scalebar'] = False
    fg, ax = plt.subplots(1, 1)

    ####
    # plot arcs of the rings
    ####
    for i in range(len(d)):
        pac = patches.Arc((0, 0), d[i] * 2, d[i] * 2, angle=0, theta1=45, theta2=360, color=tags['ring color'])
        ax.add_patch(pac)

    ####
    # show image in background
    ####
    if 'plot image' in tags:
        l = -tags['plot image FOV'] / 2 + tags['plot shift x']
        r = tags['plot image FOV'] / 2 + tags['plot shift x']
        t = -tags['plot image FOV'] / 2 + tags['plot shift y']
        b = tags['plot image FOV'] / 2 + tags['plot shift y']
        if 'plot image FOV Y' in tags:
            t = -tags['plot image FOV Y'] / 2 + tags['plot shift y']
            b = tags['plot image FOV Y'] / 2 + tags['plot shift y']

        plt.imshow(tags['plot image'], extent=(l, r, t, b), cmap='gray')

    ax.set_aspect("equal")

    # fg.canvas.draw()

    if tags['plot scalebar']:
        def f(axis):
            l = axis.get_majorticklocs()
            return len(l) > 1 and (l[1] - l[0])

        sizex = f(ax.xaxis)
        labelx = str(sizex) + ' 1/nm'
        scalebar = AnchoredSizeBar(ax.transData, sizex, labelx, loc=3,
                                   pad=0.5, color='white', frameon=False)
        # size_vertical=.2, fill_bar = True) # will be implented in matplotlib 2.1

        ax.add_artist(scalebar)
        ax.axis('off')

    #####
    # plot profile
    #####

    y = tags['Ring_Pattern']['profile_y']
    y = y / y.max() * tags['profile height']
    x = tags['Ring_Pattern']['profile_x']
    ax.plot(x, y, c=tags['profile color'])

    ax.plot([0, x[-1]], [0, 0], c=tags['profile color'])

    if 'experimental profile y' in tags:
        yy = tags['experimental profile y']
        yy = yy / yy.max() * tags['profile height']
        xx = tags['experimental profile x']
        ax.plot(xx, yy, c=tags['experimental profile color'])

    if 'plot image FOV' in tags:
        max_d = tags['plot image FOV'] / 2 + tags['plot shift x']
    else:
        max_d = d.max()
    for i in range(len(d)):
        if d[i] < max_d:
            plt.text(d[i] - .2, -.5, label[i], fontsize=tags['label size'], color=tags['label color'], rotation=90)

    if 'plot FOV' in tags:
        l = -tags['plot FOV'] / 2
        r = tags['plot FOV'] / 2
        t = -tags['plot FOV'] / 2
        b = tags['plot FOV'] / 2
        plt.xlim(l, r)
        plt.ylim(t, b)

    fg.show()


def plot_diffraction_pattern(tagsD, grey=False):
    # Get information from dictionary
    HOLZ = tagsD['HOLZ']
    ZOLZ = tagsD['allowed']['ZOLZ']
    Kikuchi = tagsD['Kikuchi']

    Laue_Zone = tagsD['allowed']['Laue_Zone']

    label = tagsD['allowed']['label']
    hkl_label = tagsD['allowed']['hkl']

    angle = np.radians(tagsD['plot rotation'])  # mrad
    c = np.cos(angle)
    s = np.sin(angle)
    r_mat = np.array([[c, -s, 0], [s, c, 0], [0, 0, 1]])

    # HOLZ and Kikuchi lines coordinates in Hough space
    LC = tagsD['Laue_circle']
    gd = np.dot(tagsD['HOLZ']['g deficient'] + LC, r_mat)
    ge = np.dot(tagsD['HOLZ']['g excess'] + LC, r_mat)
    points = np.dot(tagsD['allowed']['g'] + LC, r_mat)

    theta = tagsD['HOLZ']['theta'] + angle

    if 'thickness' not in tagsD:
        tagsD['thickness'] = 0.
    if tagsD['thickness'] > 0.1:
        intensity = np.real(tagsD['allowed']['Ig'])
    else:
        intensity = tagsD['allowed']['intensities']

    radius = tagsD['convergence_angle_nm-1']

    if radius < 0.1:
        radiusI = 2
    else:
        radiusI = radius
    # Beginning and ends of HOLZ lines
    maxlength = radiusI * 1.3
    Hxp = gd[:, 0] + maxlength * np.cos(np.pi - theta)
    Hyp = gd[:, 1] + maxlength * np.sin(np.pi - theta)
    Hxm = gd[:, 0] - maxlength * np.cos(np.pi - theta)
    Hym = gd[:, 1] - maxlength * np.sin(np.pi - theta)

    # Beginning and ends of excess HOLZ lines
    maxlength = radiusI * .8
    Exp = ge[:, 0] + maxlength * np.cos(np.pi - theta)
    Eyp = ge[:, 1] + maxlength * np.sin(np.pi - theta)
    Exm = ge[:, 0] - maxlength * np.cos(np.pi - theta)
    Eym = ge[:, 1] - maxlength * np.sin(np.pi - theta)

    # Beginning and ends of HOLZ lines
    maxlength = 20
    Kxp = gd[:, 0] + maxlength * np.cos(np.pi - theta)
    Kyp = gd[:, 1] + maxlength * np.sin(np.pi - theta)
    Kxm = gd[:, 0] - maxlength * np.cos(np.pi - theta)
    Kym = gd[:, 1] - maxlength * np.sin(np.pi - theta)

    if tagsD['linewidth Kikuchi'] < 0:
        if len(intensity[ZOLZ]) > 0:
            intensity_Kikuchi = intensity * 4. / intensity[ZOLZ].max()
        else:
            intensity_Kikuchi = intensity
    else:
        intensity_Kikuchi = np.ones(len(intensity)) * tagsD['linewidth Kikuchi']

    if tagsD['linewidth HOLZ'] < 0:
        intensity_HOLZ = np.log(intensity + 1)

        if tagsD['HOLZ']['HOLZ'].any():
            pass  # intensity_HOLZ = intensity/intensity[tagsD['HOLZ']['HOLZ']].max()*4.
    else:
        intensity_HOLZ = np.ones(len(intensity)) * tagsD['linewidth HOLZ']

    ########
    # Plot #
    ########
    cms = mpl.cm
    # cm = cms.plasma#jet#, cms.gray, cms.autumn]
    cm = plt.get_cmap(tagsD['color map'])

    # fig = plt.figure()
    fig = plt.figure()

    ax = plt.gca()
    if 'background' not in tagsD:
        tagsD['background'] = None
    if tagsD['background'] is not None:
        ax.set_facecolor(tagsD['background'])

    if 'plot image' in tagsD:
        l = -tagsD['plot image FOV'] / 2 + tagsD['plot shift x']
        r = tagsD['plot image FOV'] / 2 + tagsD['plot shift x']
        t = -tagsD['plot image FOV'] / 2 + tagsD['plot shift y']
        b = tagsD['plot image FOV'] / 2 + tagsD['plot shift y']
        if 'plot image FOV Y' in tagsD:
            t = -tagsD['plot image FOV Y'] / 2 + tagsD['plot shift y']
            b = tagsD['plot image FOV Y'] / 2 + tagsD['plot shift y']

        plt.imshow(tagsD['plot image'], extent=(l, r, t, b))

    ix = np.argsort((points ** 2).sum(axis=1))
    p = points[ix]
    inten = intensity[ix]
    reflection = hkl_label[ix]
    Lauecolor = []

    labelP = ''
    lineLabel = []

    def onpick(event):
        if isinstance(event.artist, Line2D):
            thisline = event.artist
            ind = ax.lines.index(thisline)
            print(ind, len(points), ind - len(points))
            # ind = ind- len(points)
            h, k, l = lineLabel[ind]

            if Laue_Zone[ind] > 0:
                labelP = 'Laue Zone %1d; HOLZ line: [%1d,%1d,%1d]' % (Laue_Zone[ind], h, k, l)
            else:
                labelP = 'Kikuchi line: [%1d,%1d,%1d]' % (h, k, l)
            # print(labelP)

        elif isinstance(event.artist, Circle):
            print('Circle')

        else:
            ind = event.ind[0]
            h, k, l = reflection[ind]

            print('Reflection: [%1d,%1d,%1d]' % (h, k, l))

    for i in range(int(Laue_Zone.max()) + 1):
        if i < len(tagsD['color Laue Zones']):
            Lauecolor.append(tagsD['color Laue Zones'][i])
        else:
            Lauecolor.append(tagsD['color Laue Zones'][-1])

    if 'plot_labels' not in tagsD:
        tagsD['plot_labels'] = True
    if tagsD['plot reflections']:
        print(radius)
        if radius < 0.1:
            if tagsD['color reflections'] == 'intensity':
                for i in range(len(points)):
                    ax.scatter(points[i, 0], points[i, 1], c=np.log(intensity[i] + 1), cmap=cm, s=100)

                    if tagsD['plot_labels']:
                            plt.text(points[i, 0], points[i, 1], label[i], fontsize=10)
                    #print(label[i])

            else:
                for i in range(len(Laue_Zone)):
                    color = Lauecolor[int(Laue_Zone[i])]
                    ax.scatter(points[i, 0], points[i, 1], c=color, cmap=cm, s=100)
                    if tagsD['plot_labels']:
                        plt.text(points[i, 0], points[i, 1], label[i], fontsize=8)
                    #print(label[i])

            ax.scatter(0, 0, c=tagsD['color zero'], s=100)
            radius = 2
        else:
            ix = np.argsort((points ** 2).sum(axis=1))
            p = points[ix]
            inten = intensity[ix]
            if tagsD['color reflections'] == 'intensity':
                circles(p[:, 0], p[:, 1], s=radius, c=np.log(inten + 1), cmap=cm, alpha=0.9, edgecolor='', picker=5)  #
            else:
                for i in range(len(Laue_Zone)):
                    color = Lauecolor[int(Laue_Zone[i])]
                    circles(p[i, 0], p[i, 1], s=radius, c=color, cmap=cm, alpha=0.9, edgecolor='', picker=5)  #
                    plt.text(points[i, 0], points[i, 1], label[i], fontsize=8)

    k = 0
    if tagsD['plot HOLZ']:
        for i in range(len(Hxp)):

            if tagsD['HOLZ']['HOLZ'][i]:
                color = Lauecolor[int(Laue_Zone[i])]
                if tagsD['plot HOLZ']:
                    # plot HOLZ lines
                    line, = plt.plot((Hxp[i], Hxm[i]), (Hyp[i], Hym[i]), c=color, linewidth=intensity_HOLZ[i], picker=5)
                    if tagsD['label HOLZ']:  ## Add indices
                        plt.text(Hxp[i], Hyp[i], label[i], fontsize=8)
                    lineLabel.append(hkl_label[i])
                    # print(i, hkl_label[i], intensity_HOLZ[i])

                if tagsD['plot HOLZ excess']:
                    line, = plt.plot((Exp[i], Exm[i]), (Eyp[i], Eym[i]), c=color, linewidth=intensity_HOLZ[i])
                    lineLabel.append(hkl_label[i])

                    if tagsD['label HOLZ']:  ## Add indices
                        plt.text(Exp[i], Eyp[i], label[i], fontsize=8)


            else:
                # Plot Kikuchi lines
                if tagsD['plot Kikuchi']:
                    line, = plt.plot((Kxp[i], Kxm[i]), (Kyp[i], Kym[i]), c=tagsD['color Kikuchi'],
                                     linewidth=intensity_Kikuchi[i] / 10, picker=5)
                    if tagsD['label Kikuchi']:  # Add indices

                        plt.text(Kxp[i], Kyp[i], label[i], fontsize=tagsD['label size'], color=tagsD['label color'])
                    lineLabel.append(hkl_label[i])
    elif tagsD['plot Kikuchi']:
        # Beginning and ends of Kikuchi lines
        maxlength = tagsD['plot FOV']
        gd = tagsD['Kikuchi']['min dist']
        theta = tagsD['Kikuchi']['theta']
        Kxp = gd[:, 0] + maxlength * np.cos(np.pi - theta)
        Kyp = gd[:, 1] + maxlength * np.sin(np.pi - theta)
        Kxm = gd[:, 0] - maxlength * np.cos(np.pi - theta)
        Kym = gd[:, 1] - maxlength * np.sin(np.pi - theta)
        for i in range(len(Kxp)):
            line, = plt.plot((Kxp[i], Kxm[i]), (Kyp[i], Kym[i]), c=tagsD['color Kikuchi'], linewidth=2)

    def format_coord(x, y):
        return labelP + 'x=%1.4f, y=%1.4f' % (x, y)

    ax.format_coord = format_coord

    if not (tagsD['color ring zero'] == 'None'):
        ring = plt.Circle((0, 0), radius, color=tagsD['color ring zero'], fill=False, linewidth=2)
        ax.add_artist(ring)
        # print(ring)
    if not tagsD['color zero'] == 'None':
        circle = plt.Circle((0, 0), radius, color=tagsD['color zero'], linewidth=2)
        ax.add_artist(circle)

    plt.axis('equal')
    if 'plot FOV' in tagsD:
        l = -tagsD['plot FOV'] / 2
        r = tagsD['plot FOV'] / 2
        t = -tagsD['plot FOV'] / 2
        b = tagsD['plot FOV'] / 2
        plt.xlim(l, r)
        plt.ylim(t, b)

    fig.canvas.mpl_connect('pick_event', onpick)
    # texfig.savefig("HOLZ")

    # plt.title( tagsD['crystal'])
    plt.show()
